# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.utils.configuration_utils import get_hostname
from contrast.utils.configuration_utils import str_to_bool


class Server(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="server.name",
                default_value=get_hostname(),
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="server.discover_cloud_resource",
                default_value=True,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="server.path", default_value="/", type_cast=str
            ),
            ConfigOption(canonical_name="server.type", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="server.version", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="server.environment", default_value="", type_cast=str
            ),
            ConfigOption(canonical_name="server.tags", default_value="", type_cast=str),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from dataclasses import dataclass, field
from contrast.api.user_input import UserInput
from contrast.utils.timer import now_ms


@dataclass
class Sample:
    user_input: UserInput
    stack: list
    details: dict = field(default_factory=dict)
    timestamp_ms: int = field(default_factory=now_ms)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.utils.configuration_utils import str_to_bool
from contrast.utils.loggers import (
    DEFAULT_SECURITY_LOG_PATH,
    DEFAULT_SECURITY_LOG_LEVEL,
    DEFAULT_LOG_PATH,
    DEFAULT_LOG_LEVEL,
)


class Agent(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            # Some logger default values are handled by the logger
            ConfigOption(
                canonical_name="agent.logger.level",
                default_value=DEFAULT_LOG_LEVEL,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.logger.path",
                default_value=DEFAULT_LOG_PATH,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.logger.stdout",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.logger.stderr",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.logger.backups",
                default_value=10,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.logger.roll_size",
                default_value=100,  # specified in MB
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.path",
                default_value=DEFAULT_SECURITY_LOG_PATH,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.level",
                default_value=DEFAULT_SECURITY_LOG_LEVEL,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.protocol",
                default_value="UDP",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.server_host",
                default_value=None,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.ip",
                default_value=None,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.port",
                default_value=None,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.facility",
                default_value=19,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.severity_exploited",
                default_value="ALERT",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.severity_blocked",
                default_value="NOTICE",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.severity_blocked_perimeter",
                default_value="NOTICE",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.severity_probed",
                default_value="WARNING",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.security_logger.syslog.severity_suspicious",
                default_value="WARNING",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="agent.python.enable_automatic_middleware",
                default_value=True,
                type_cast=str_to_bool,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="agent.python.enable_drf_response_analysis",
                default_value=True,
                type_cast=str_to_bool,
                log_effective_config=False,
            ),
            # TODO: PYT-3337 - we should issue a deprecation warning if
            # agent.python.enable_profiler is used
            ConfigOption(
                canonical_name="agent.python.enable_profiler",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.python.profiler.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.python.tracer.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="agent.python.tracer.min_duration_ms",
                default_value=0.010,  # 10us
                type_cast=float,
            ),
            ConfigOption(
                canonical_name="agent.polling.app_settings_ms",
                default_value=30_000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.polling.server_settings_ms",
                default_value=30_000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.polling.heartbeat_ms",
                default_value=30_000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="agent.route_coverage.report_on_error",
                default_value=False,
                type_cast=str_to_bool,
            ),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
]


untrusted_deserialization_triggers = [
    {
        "module": "marshal",
        "method_name": ["load", "loads"],
        # does not accept kwargs
        "source": "ARG_0",
        "protect_mode": True,
    },
    {
        "module": "pickle",
        "method_name": "load",
        "source": "ARG_0,KWARG:file",
        "protect_mode": True,
    },
    {
        "module": "pickle",
        "method_name": "loads",
        "source": "ARG_0,KWARG:data",
        "protect_mode": True,
    },
    {
        "module": "yaml.loader",
        "class_name": "BaseLoader",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:stream",
        "protect_mode": True,
    },
    {
        "module": "yaml.loader",
        "class_name": ["Loader", "UnsafeLoader", "FullLoader"],
        "method_name": "__init__",
        "source": "ARG_0,KWARG:stream",
        "protect_mode": True,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "untrusted-deserialization",
        untrusted_deserialization_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

from __future__ import annotations

import sys
import traceback

from io import StringIO
from types import FrameType
from typing import Callable

from .contextvars import _ASYNC_CALLING_STACK
from .typing import ExcInfo


def _format_exception(exc_info: ExcInfo) -> str:
    """
    Prettyprint an `exc_info` tuple.

    Shamelessly stolen from stdlib's logging module.
    """
    if exc_info == (None, None, None):  # type: ignore[comparison-overlap]
        return "MISSING"

    sio = StringIO()

    traceback.print_exception(exc_info[0], exc_info[1], exc_info[2], None, sio)
    s = sio.getvalue()
    sio.close()
    if s[-1:] == "\n":
        s = s[:-1]

    return s


def _find_first_app_frame_and_name(
    additional_ignores: list[str] | None = None,
    *,
    _getframe: Callable[[], FrameType] = sys._getframe,
) -> tuple[FrameType, str]:
    """
    Remove all intra-structlog calls and return the relevant app frame.

    Args:
        additional_ignores:
            Additional names with which the first frame must not start.

        _getframe:
            Callable to find current frame. Only for testing to avoid
            monkeypatching of sys._getframe.

    Returns:
        tuple of (frame, name)
    """
    ignores = tuple(["structlog"] + (additional_ignores or []))
    f = _ASYNC_CALLING_STACK.get(_getframe())
    name = f.f_globals.get("__name__") or "?"
    while name.startswith(ignores):
        if f.f_back is None:
            name = "?"
            break
        f = f.f_back
        name = f.f_globals.get("__name__") or "?"
    return f, name


def _format_stack(frame: FrameType) -> str:
    """
    Pretty-print the stack of *frame* like logging would.
    """
    sio = StringIO()

    sio.write("Stack (most recent call last):\n")
    traceback.print_stack(frame, file=sio)
    sinfo = sio.getvalue()
    if sinfo[-1] == "\n":
        sinfo = sinfo[:-1]
    sio.close()

    return sinfo



from __future__ import annotations

import os
import pkgutil
import sys
from functools import lru_cache

long_versions = [
    "2.6.9",
    "2.7.9",
    "3.2.6",
    "3.3.6",
    "3.4.3",
    "3.5",
    "3.6",
    "3.7",
    "3.8",
    "3.9",
    "3.10",
    "3.11",
    "3.12",
    "3.13",
]

short_versions = [".".join(x.split(".")[:2]) for x in long_versions]


def get_canonical_version(version: str) -> str:
    if version in long_versions:
        version = ".".join(version.split(".")[:2])
    elif version not in short_versions:
        raise ValueError(f"No such version: {version}")

    return version


def stdlib_list(version: str | None = None) -> list[str]:
    """
    Given a ``version``, return a ``list`` of names of the Python Standard
    Libraries for that version.

    :param str|None version: The version (as a string) whose list of libraries you want
        (formatted as ``X.Y``, e.g. ``"2.7"`` or ``"3.10"``).

        If not specified, the current version of Python will be used.

    :return: A list of standard libraries from the specified version of Python
    :rtype: list
    """

    version = (
        get_canonical_version(version)
        if version is not None
        else ".".join(str(x) for x in sys.version_info[:2])
    )

    module_list_file = os.path.join("lists", f"{version}.txt")

    data = pkgutil.get_data("contrast_vendor.stdlib_list", module_list_file).decode()  # type: ignore[union-attr]

    result = [y for y in [x.strip() for x in data.splitlines()] if y]

    return result


@lru_cache(maxsize=16)
def _stdlib_list_with_cache(version: str | None = None) -> list[str]:
    """Internal cached version of `stdlib_list`"""
    return stdlib_list(version=version)


@lru_cache(maxsize=256)
def in_stdlib(module_name: str, version: str | None = None) -> bool:
    """
    Return a ``bool`` indicating if module ``module_name`` is in the list of stdlib
    symbols for python version ``version``. If ``version`` is ``None`` (default), the
    version of current python interpreter is used.

    Note that ``True`` will be returned for built-in modules too, since this project
    considers they are part of stdlib. See :issue:21.

    It relies on ``@lru_cache`` to cache the stdlib list and query results for similar
    calls. Therefore it is much more efficient than ``module_name in stdlib_list()``
    especially if you wish to perform multiple checks.

    :param str|None module_name: The module name (as a string) to query for.
    :param str|None version: The version (as a string) whose list of libraries you want
        (formatted as ``X.Y``, e.g. ``"2.7"`` or ``"3.10"``).

        If not specified, the current version of Python will be used.

    :return: A bool indicating if the given module name is part of standard libraries
        for the specified version of Python.
    :rtype: list
    """
    ref_list = _stdlib_list_with_cache(version=version)
    return module_name in ref_list



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .agent import Agent
from .api import Api
from .application import Application
from .assess import Assess
from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from .agent_config import AgentConfig
from .inventory import Inventory
from .observe import Observe
from .protect import Protect
from .reporting import Reporting
from .root import Root
from .server import Server

__all__ = [
    "Agent",
    "Api",
    "Application",
    "Assess",
    "ConfigBuilder",
    "ConfigOption",
    "AgentConfig",
    "Inventory",
    "Observe",
    "Protect",
    "Reporting",
    "Root",
    "Server",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent.assess.utils import copy_from
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.utils.string_utils import ensure_string


class ReprPropagator(BasePropagator):
    """
    Propagation method for __repr__ method of stringy types

    If the string itself is faithfully represented in the the repr, then we simply copy
    the tags to the correct offset to account for boilerplate. Otherwise, we just splat
    across the result.
    """

    def _propagate(self):
        # Get a stringy version of the source
        if isinstance(self.first_source, bytearray):
            string = ensure_string(bytes(self.first_source), errors="ignore")
        else:
            string = ensure_string(self.first_source, errors="ignore")

        # See whether the stringy version is inside the repr
        offset = self.target.find(string)
        # Handle edge case where the string itself is "bytearray" or something like that
        if offset == 0:
            offset = self.target.rfind(string)

        if offset >= 0:
            copy_from(self.target, self.first_source, offset, self.node.untags)
            return

        # If it's not, then just splat to account for encoding, etc.
        self.splat_tags(self.first_source, self.target)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections import defaultdict
from functools import cached_property
import hashlib
import os
from pathlib import Path
import queue
import threading
import time
import uuid
import sys

import contrast

from contrast.agent import agent_state
from contrast import __version__
from contrast.agent import scope
from contrast import AGENT_CURR_WORKING_DIR
from contrast.agent.settings import Settings
from contrast_vendor import structlog as logging
from requests import post as post_request, get as get_request
from contrast.agent.events import StartupMetricsTelemetryEvent, ErrorTelemetryEvent

logger = logging.getLogger("contrast")

FILE_LOCATIONS = [
    # Most stable place but some OS may not allow creating here
    "/etc/contrast/python/",
    os.path.join(AGENT_CURR_WORKING_DIR, "config", "contrast"),
]

DISCLAIMER = (
    "The Contrast Python Agent collects usage data "
    "in order to help us improve compatibility and security coverage. "
    "The data is anonymous and does not contain application data. "
    "It is collected by Contrast and is never shared. "
    "You can opt-out of telemetry by setting the "
    "CONTRAST_AGENT_TELEMETRY_OPTOUT environment variable to '1' or 'true'. "
    "Read more about Contrast Python Agent telemetry: "
    "https://docs.contrastsecurity.com/en/python-telemetry.html"
)

BASE_URL = "https://telemetry.python.contrastsecurity.com"
BASE_ENDPOINT = "api/v1/telemetry"
URL = f"{BASE_URL}/{BASE_ENDPOINT}"
HEADERS = {"User-Agent": f"python-{__version__}"}

TELEMETRY_THREAD_NAME = "ContrastTelemetry"


class Telemetry(threading.Thread):
    # While the telemetry spec suggests waiting for 3 hours, we've decided
    # that 30 minutes is reasonable for this agent.
    WAIT = 60 * 30  # 30 mins
    RETRY_WAIT = 60
    QUEUE_GET_TIMEOUT = 5
    REPORTED_ERRORS = {}
    # Limit the number of messages stored each reporting period.
    MAX_MESSAGES = 300

    def __init__(self):
        super().__init__()
        self.daemon = True
        self.name = TELEMETRY_THREAD_NAME

        self.enabled = True
        self.is_public_build = True
        self.message_q = queue.Queue(maxsize=100)
        self._stopper = object()  # sentinel sent over message_q to stop the thread
        self._msgs = defaultdict(list)
        self._msg_count = 0
        self.settings = Settings()

    def stop(self):
        self.message_q.put(self._stopper)
        timeout = 30
        self.join(timeout)
        if self.is_alive():
            logger.warn("failed to stop Telemetry thread", timeout=timeout)

    @cached_property
    def instance_id(self):
        if self._mac_addr is None:
            return "_" + uuid.uuid4().hex
        return self._sha256(hex(self._mac_addr))

    @cached_property
    def application_id(self):
        if self._mac_addr is None:
            return "_" + uuid.uuid4().hex
        return self._sha256(hex(self._mac_addr) + agent_state.get_app_name())

    @cached_property
    def _mac_addr(self):
        """
        The MAC address for the current machine's primary network adapter as a base-10
        integer. If we find a multicast MAC address, return None.
        See _is_multicast_mac_address.
        """
        _mac_addr = uuid.getnode()
        if self._is_multicast_mac_address(_mac_addr):
            return None
        return _mac_addr

    def run(self):
        self._check_enabled()

        if not self.enabled:
            return

        # Ensure thread runs in scope because it is initialized
        # before our thread.start patch is applied.
        with scope.contrast_scope():
            logger.debug("Starting telemetry thread")

            # Do not move creating startup msg outside of this function
            # so the work stays in the telemetry thread, not the main thread.
            startup_msg = StartupMetricsTelemetryEvent(self.instance_id)
            logger.debug("Sending StartupMetrics msg: %s", startup_msg)
            self._store_message(startup_msg)
            self._send_messages()  # send StartupMetrics immediately

            # send first message immediately, future messages will be sent after wait_time.
            next_send_time = time.time()
            while self.settings.is_agent_config_enabled():
                try:
                    msg = self.message_q.get(block=True, timeout=self.QUEUE_GET_TIMEOUT)
                    if msg is self._stopper:
                        self._send_messages()
                        break
                    self._store_message(msg)
                    now = time.time()
                    if now >= next_send_time:
                        wait_time = self._send_messages()
                        next_send_time = now + wait_time
                except queue.Empty:
                    pass
                except Exception as e:
                    logger.debug(
                        "WARNING: reporting client failed to send message", exc_info=e
                    )

    def add_message(self, msg):
        if not self.enabled or msg is None:
            return

        logger.debug("Adding msg to telemetry queue: %s", msg)
        try:
            self.message_q.put_nowait(msg)
        except queue.Full:
            logger.debug("WARNING: Telemetry queue is full, dropping message", msg=msg)

    def _send_messages(self) -> int:
        """
        Send all stored messages, one batch request for each path for group of
        messages. Returns a time in seconds that a caller should wait before
        sending messages again.
        """
        # It would be more natural to iterate over self._msgs.items() here, but
        # if a retry_after response is sent, we can't easily remove already sent
        # messages from self._msgs or save still-to-be-sent messages from within
        # the items() loop. Instead, we pop individual messages to handle this
        # case.
        for _ in range(len(self._msgs)):
            path, messages = self._msgs.popitem()
            self._msg_count -= len(messages)
            try:
                response = self._post(messages, path)
                if retry_wait := self._retry_after(response):
                    # readd message so it can be sent again after retry_wait.
                    self._msgs[path] = messages
                    return retry_wait
            except Exception as ex:
                logger.debug(
                    "Could not send batch of telemetry messages.", exception=str(ex)
                )

        return self.WAIT

    def _store_message(self, msg):
        if self._msg_count >= self.MAX_MESSAGES:
            logger.debug(
                "WARNING: message store limit reached, dropping message", msg=msg
            )
            return
        self._msgs[msg.path].append(msg)
        self._msg_count += 1

    def should_report_error(self, error, original_func):
        key = " ".join(
            [
                type(error).__name__,
                original_func.__name__,
                original_func.__module__,
                str(error),
            ]
        )

        if key in self.REPORTED_ERRORS:
            self.REPORTED_ERRORS[key] += 1
            return False

        self.REPORTED_ERRORS[key] = 1

        return True

    def report_error(self, error, original_func, logger_="", message="", skip_frames=0):
        """
        Report an agent error/exception to Telemetry.

        Take great care to avoid calling this where application errors or customer
        code may be caught.
        """
        if self.should_report_error(error, original_func):
            self.add_message(
                ErrorTelemetryEvent(
                    self.instance_id,
                    error=error,
                    logger_=logger_,
                    message=message,
                    # +1 to remove the current report_error frame
                    skip_frames=skip_frames + 1,
                )
            )

    def _post(self, messages, path):
        """
        Send a list of `messages` to telemetry `path`
        """
        logger.debug("Sending %s Telemetry messages to %s.", len(messages), path)

        response = post_request(
            f"{URL}{path}",
            json=[msg.to_json() for msg in messages],
            headers=HEADERS,
            allow_redirects=False,
            verify=True,
        )

        logger.debug("Telemetry response: %s %s", response.status_code, response.reason)
        return response

    def _retry_after(self, response) -> int | None:
        """
        Per RFC-6585, check response status code for 429 and returns the value
        of the Retry-After present or self.RETRY_WAIT.
        """
        if response.status_code == 429:
            wait_time = int(response.headers.get("Retry-After", self.RETRY_WAIT))
            logger.debug("Telemetry waiting for %s seconds", wait_time)
            return wait_time
        return None

    def _is_multicast_mac_address(self, mac_addr):
        """
        A multicast MAC address is an indication that we're not seeing a hardware MAC
        address, which means this value is subject to change even on a single server.
        MAC addresses have a multicast bit that is only set for such addresses. This
        method returns True if the supplied mac address is a multicast address.

        Note that when uuid.getnode() isn't able to find a hardware MAC address, it
        randomly generates an address and (critically) sets the multicast bit.
        """
        return bool(mac_addr & (1 << 40))

    def _sha256(self, str_input):
        return hashlib.sha256(str_input.encode()).hexdigest()

    def _check_enabled(self):
        self._check_is_public_build()

        if contrast.telemetry_disabled() or self._connection_failed():
            self.enabled = False
        else:
            self._find_or_create_file()

        # Debug log for dev purposes. The only time an agent user should see anything
        # about telemetry is if the disclaimer is print/logged.
        logger.debug("Agent telemetry is %s", "enabled" if self.enabled else "disabled")

    def _check_is_public_build(self) -> None:
        is_public = os.environ.get("CONTRAST_PUBLIC_BUILD")
        self.is_public_build = True

        if is_public and is_public.lower() in ("0", "false"):
            self.is_public_build = False

        # Debug log for dev purposes. The only time an agent user should see anything
        # about telemetry is if the disclaimer is print/logged.
        logger.debug(
            "Agent telemetry %s",
            (
                "is in public build mode"
                if self.is_public_build
                else "is not in public build mode"
            ),
        )

    def _connection_failed(self):
        try:
            # any response here is fine as long as no error is raised.
            get_request(BASE_URL)
            return False
        except Exception as ex:
            # Any exception such as SSLError, ConnectionError, etc
            logger.debug("Telemetry connection failed: %s", ex)

        return True

    def _find_or_create_file(self):
        """
        Find an existing .telemetry file or create an empty one.

        /etc/contrast/python/ is the preferred location because it's permanent
        across any agent, but in some OS we may not be able to create it.

        The .telemetry file is intended to be an empty file only as a marker
        to let us know if we have print/logged the disclaimer. Failing to find it
        in any situation means we should print/log.
        """
        name = ".telemetry"

        # 1. If .telemetry file exists, don't print/log disclaimer
        for path in FILE_LOCATIONS:
            file_path = os.path.join(path, name)
            if Path(file_path).exists():
                return

        # 2. If .telemetry file does not exist, attempt to create dir structure
        # and the empty file
        for path in FILE_LOCATIONS:
            file_path = os.path.join(path, name)
            try:
                Path(path).mkdir(parents=True, exist_ok=True)
                Path(file_path).touch()
                break
            except Exception:
                continue

        # 3. Print/log disclaimer if .telemetry file was created or if it failed to
        # be created
        sys.stderr.write(f"{DISCLAIMER}\n")  # pylint: disable=superfluous-parens
        logger.info(DISCLAIMER)



import re

__all__ = ['Range', 'ContentRange']
_rx_range = re.compile(r'bytes *= *(\d*) *- *(\d*)', flags=re.I)
_rx_content_range = re.compile(r'bytes (?:(\d+)-(\d+)|[*])/(?:(\d+)|[*])')

class Range(object):
    """
        Represents the Range header.
    """

    def __init__(self, start, end):
        assert end is None or end >= 0, "Bad range end: %r" % end
        self.start = start
        self.end = end # non-inclusive

    def range_for_length(self, length):
        """
            *If* there is only one range, and *if* it is satisfiable by
            the given length, then return a (start, end) non-inclusive range
            of bytes to serve.  Otherwise return None
        """
        if length is None:
            return None
        start, end = self.start, self.end
        if end is None:
            end = length
            if start < 0:
                start += length
        if _is_content_range_valid(start, end, length):
            stop = min(end, length)
            return (start, stop)
        else:
            return None

    def content_range(self, length):
        """
            Works like range_for_length; returns None or a ContentRange object

            You can use it like::

                response.content_range = req.range.content_range(response.content_length)

            Though it's still up to you to actually serve that content range!
        """
        range = self.range_for_length(length)
        if range is None:
            return None
        return ContentRange(range[0], range[1], length)

    def __str__(self):
        s,e = self.start, self.end
        if e is None:
            r = 'bytes=%s' % s
            if s >= 0:
                r += '-'
            return r
        return 'bytes=%s-%s' % (s, e-1)

    def __repr__(self):
        return '<%s bytes %r-%r>' % (
            self.__class__.__name__,
            self.start, self.end)

    def __iter__(self):
        return iter((self.start, self.end))

    @classmethod
    def parse(cls, header):
        """
            Parse the header; may return None if header is invalid
        """
        m = _rx_range.match(header or '')
        if not m:
            return None
        start, end = m.groups()
        if not start:
            return cls(-int(end), None)
        start = int(start)
        if not end:
            return cls(start, None)
        end = int(end) + 1 # return val is non-inclusive
        if start >= end:
            return None
        return cls(start, end)


class ContentRange(object):

    """
    Represents the Content-Range header

    This header is ``start-stop/length``, where start-stop and length
    can be ``*`` (represented as None in the attributes).
    """

    def __init__(self, start, stop, length):
        if not _is_content_range_valid(start, stop, length):
            raise ValueError(
                "Bad start:stop/length: %r-%r/%r" % (start, stop, length))
        self.start = start
        self.stop = stop # this is python-style range end (non-inclusive)
        self.length = length

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)

    def __str__(self):
        if self.length is None:
            length = '*'
        else:
            length = self.length
        if self.start is None:
            assert self.stop is None
            return 'bytes */%s' % length
        stop = self.stop - 1 # from non-inclusive to HTTP-style
        return 'bytes %s-%s/%s' % (self.start, stop, length)

    def __iter__(self):
        """
            Mostly so you can unpack this, like:

                start, stop, length = res.content_range
        """
        return iter([self.start, self.stop, self.length])

    @classmethod
    def parse(cls, value):
        """
            Parse the header.  May return None if it cannot parse.
        """
        m = _rx_content_range.match(value or '')
        if not m:
            return None
        s, e, l = m.groups()
        if s:
            s = int(s)
            e = int(e) + 1
        l = l and int(l)
        if not _is_content_range_valid(s, e, l, response=True):
            return None
        return cls(s, e, l)


def _is_content_range_valid(start, stop, length, response=False):
    if (start is None) != (stop is None):
        return False
    elif start is None:
        return length is None or length >= 0
    elif length is None:
        return 0 <= start < stop
    elif start >= stop:
        return False
    elif response and stop > length:
        # "content-range: bytes 0-50/10" is invalid for a response
        # "range: bytes 0-50" is valid for a request to a 10-bytes entity
        return False
    else:
        return 0 <= start < length



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_header_only_rule import (
    BaseHeaderOnlyRule,
)


class ClickjackingRule(BaseHeaderOnlyRule):
    """
    Rule is violated if the header X-Frame-Options is missing or present
    but not assigned to a desired value.
    """

    @property
    def name(self):
        return "clickjacking-control-missing"

    @property
    def header_key(self):
        return "X-Frame-Options"

    @property
    def good_values(self):
        return ("deny", "sameorigin")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.trigger_rule import TriggerRule
from contrast_fireball import AssessEvent


class DataflowRule(TriggerRule):
    """
    Rule class used for all dataflow rules. Hash computation includes events.
    """

    def update_preflight_hash(self, hasher, events: list[AssessEvent] = None, **kwargs):
        """
        This gets called from `create_finding` in the base class

        It turns out that only three things differentiate dataflow findings:
            1. The name of the rule that was triggered
            2. The name and type of the source event(s)
            3. The request context

        Information from the current request is used to update the hash at the
        end of the request lifecycle in middleware.
        """
        for event in events or []:
            for source in event.event_sources:
                hasher.update(source.source_type)
                hasher.update(source.source_name)

    @classmethod
    def from_nodes(cls, name, nodes, disallowed_tags=None, required_tags=None):
        return super().from_nodes(name, True, nodes, disallowed_tags, required_tags)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.asgi.middleware import ASGIMiddleware


class FastApiMiddleware(ASGIMiddleware):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, framework_name="fastapi")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    register_module_patcher,
    repatch_module,
    unregister_module_patcher,
)

VENDOR = "SQLite3"


class Sqlite3Patcher(dbapi2.Dbapi2Patcher):
    @fail_quietly("failed to get database inventory information")
    def extract_connection_attributes(self, connection, connect_args, connect_kwargs):
        # sqlite does not use a server, so no need for host/port
        database = connect_kwargs.get("database") or (
            connect_args[0] if len(connect_args) > 0 else "unknown"
        )
        # this can sometimes be a PosixPath, so we need to cast
        self.db_name = str(database)
        super().extract_connection_attributes(connection, connect_args, connect_kwargs)


def instrument_sqlite3(sqlite3):
    dbapi2.instrument_adapter(
        sqlite3,
        VENDOR,
        Sqlite3Patcher,
        extra_cursors=[
            sqlite3.Cursor,
            # sqlite3.Connection.execute* functions are non-standard convenience
            # functions that will create new cursors automatically. sqlite3.Connection
            # exposes the same execute* API as a standard cursor, however, so for
            # patching purposes we can just identify it as an extra cursor. This only
            # became necessary in newer versions of python (3.11+) where the
            # implementation of these functions was moved to C.
            sqlite3.Connection,
        ],
    )


def register_patches():
    register_module_patcher(instrument_sqlite3, "sqlite3.dbapi2")
    register_module_patcher(repatch_module, "sqlite3")


def reverse_patches():
    for sqlite in ("sqlite3", "sqlite3.dbapi2"):
        unregister_module_patcher(sqlite)
        if sqlite3 := sys.modules.get(sqlite):
            patch_manager.reverse_patches_by_owner(sqlite3)
            patch_manager.reverse_patches_by_owner(sqlite3.Cursor)
            patch_manager.reverse_patches_by_owner(sqlite3.Connection)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import (
    copy_events,
    copy_tags_to_offset,
    get_properties,
)
from contrast.utils.patch_utils import get_arg


class ReplacePropagator(BasePropagator):
    def _copy_span_tags_to_offset(self, orig_properties, start_idx, end_idx, offset):
        old_span = AdjustedSpan(start_idx, end_idx)
        source_tags = orig_properties.tags_at_range(old_span)
        copy_tags_to_offset(self.target_properties, source_tags, offset)

    def _propagate_tags(
        self, orig_properties, new_properties, orig_str, old_str, new_str
    ):
        source_offset = 0
        target_offset = 0

        count = orig_str.count(old_str)
        if (
            count_arg := get_arg(
                self.preshift.args, self.preshift.kwargs, 2, "count", default=-1
            )
            > -1
        ):
            count = min(count, count_arg)

        while count > 0:
            found_idx = orig_str.find(old_str, source_offset)
            # copy old tags to the new location
            if source_offset != found_idx and orig_properties is not None:
                self._copy_span_tags_to_offset(
                    orig_properties,
                    source_offset,
                    found_idx,
                    target_offset,
                )

            target_offset += found_idx - source_offset
            source_offset = found_idx + len(old_str)

            # copy any tags from the new string
            if new_properties is not None:
                copy_tags_to_offset(
                    self.target_properties, new_properties.tags, target_offset
                )

            target_offset += len(new_str)
            count -= 1

        if source_offset < len(orig_str) and orig_properties is not None:
            self._copy_span_tags_to_offset(
                orig_properties,
                source_offset,
                len(orig_str),
                target_offset,
            )

    def _propagate(self):
        orig_str = self.preshift.obj
        old_str, new_str = self.preshift.args[:2]

        orig_properties = get_properties(orig_str)
        new_properties = get_properties(new_str)

        if orig_properties is None and new_properties is None:
            return

        self._propagate_tags(
            orig_properties, new_properties, orig_str, old_str, new_str
        )

        copy_events(self.target_properties, orig_properties)
        copy_events(self.target_properties, new_properties)

        self.target_properties.cleanup_tags()



import functools
import sys


# from jaraco.functools 4.1
def identity(x):
    return x


# from jaraco.functools 4.1
def apply(transform):
    def wrap(func):
        return functools.wraps(func)(compose(transform, func))

    return wrap


# from jaraco.functools 4.1
def compose(*funcs):
    def compose_two(f1, f2):
        return lambda *args, **kwargs: f1(f2(*args, **kwargs))

    return functools.reduce(compose_two, funcs)


def replace(pattern):
    r"""
    >>> replace(r'foo\z')
    'foo\\Z'
    """
    return pattern[:-2] + pattern[-2:].replace(r'\z', r'\Z')


legacy_end_marker = apply(replace) if sys.version_info < (3, 14) else identity



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators import BasePropagator
from contrast.agent.assess.utils import copy_events, get_properties
from contrast.utils.assess.tag_utils import merge_tags


class SlicePropagator(BasePropagator):
    def _propagate(self):
        slice_: slice = self.preshift.args[0]
        source_properties = get_properties(self.preshift.obj)

        if slice_.step in (1, None):
            # fast path for common case of step=1
            obj_len = self.preshift.obj_length
            start, stop, _ = slice_.indices(obj_len)
            span = AdjustedSpan(start, stop)
            source_tags = source_properties.tags_at_range(span)
            for label, rng in source_tags.items():
                self.target_properties.add_existing_tag(label, rng)

        else:
            # slow but correct
            #
            # This walks the source string's slice view, index by index
            # and copies each tag individually to the target string.
            # For large slices, this is slower than the fast path above.

            # This is like the coolest thing ever. I found it here:
            # https://stackoverflow.com/a/42883770/4312739
            # It converts a slice to a range of indices
            offsets = list(range(self.preshift.obj_length)[slice_])

            for target_offset, source_offset in enumerate(offsets):
                span = AdjustedSpan(source_offset, source_offset + 1)
                source_tags = source_properties.tags_at_range(span)
                for name in source_tags:
                    end_offset = target_offset + 1
                    new_span = AdjustedSpan(target_offset, end_offset)
                    self.target_properties.add_tag(name, new_span)

        copy_events(self.target_properties, source_properties)
        merge_tags(self.target_properties.tags)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy import registry
from contrast.agent.assess.policy import trigger_policy
from contrast.utils.decorators import fail_quietly


@fail_quietly("Error running unsafe code execution assess rule")
def apply_rule(module_name, method_name, result, args, kwargs):
    trigger_rule = registry.get_triggers_by_rule("unsafe-code-execution")

    trigger_nodes = trigger_rule.find_trigger_nodes(module_name, method_name)

    trigger_policy.apply(trigger_rule, trigger_nodes, result, args, kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import copy
from dataclasses import dataclass, field
from itertools import chain
from pathlib import PurePath
import re
import reprlib
import threading
import time
from typing import (
    Any,
    Callable,
)
from collections.abc import Mapping, Sequence
from contrast.agent.assess.tag import Tag

from contrast.agent.policy.constants import (
    OBJECT,
    RETURN,
)

from contrast.agent.assess.truncate import truncate_tainted_string
from contrast.agent.assess.utils import get_properties

from contrast.agent.policy.policy_node import PolicyNode, Location
from contrast.utils.assess.duck_utils import len_or_zero
from contrast.utils.base64_utils import B64_NONE_STRING, base64_encode
from contrast.utils.stack_trace_utils import (
    StackSummary,
    acceptable_frame,
    build_stack,
    to_assess_stack,
)
from contrast_fireball import (
    AssessEvent,
    AssessEventAction,
    AssessEventSource,
    AssessObject,
    AssessParentObject,
    AssessSignature,
    AssessTaintRange,
    RouteSource,
    SourceType,
)
from contrast_vendor import structlog as logging
from contrast_vendor.webob.util import html_escape

logger = logging.getLogger("contrast")

INIT = "__init__"
INITIALIZERS = (INIT, "__new__")
NONE_STRING = str(None)


class Repr(reprlib.Repr):
    def repr_str(self, x, level):
        return x

    def repr_bytes(self, x, level):
        return x.decode(errors="replace")

    def repr_bytearray(self, x, level):
        return self.repr_bytes(x, level)

    def repr_instance(self, x, level):
        return f"{x.__class__.__module__}.{x.__class__.__qualname__}@{id(x):#x}"

    def repr_bool(self, x, level):
        return str(x)

    def repr_float(self, x, level):
        return str(x)

    def repr_NoneType(self, x, level):
        return "None"

    def repr_Pattern(self, x, level):
        return (
            self.repr1(x.pattern, level)
            if isinstance(x, re.Pattern)
            else self.repr_instance(x, level)
        )

    def repr_PurePath(self, x, level):
        return str(x) if isinstance(x, PurePath) else self.repr_instance(x, level)

    repr_PosixPath = repr_WindowsPath = repr_Path = repr_PurePosixPath = (
        repr_PureWindowsPath
    ) = repr_PurePath


class FullRepr(Repr):
    def repr_StringIO(self, x, level):
        return x.getvalue()

    def repr_BytesIO(self, x, level):
        return x.getvalue().decode(errors="replace")


class TruncatedRepr:
    def __init__(self, _repr: Repr):
        _repr = copy.copy(_repr)
        _repr.maxlevel = 3
        self._repr = _repr.repr

    def repr(self, obj: object):
        s = self._repr(obj)
        if len(s) > 103:
            s = s[:50] + "..." + s[-50:]
        return s


base_repr = Repr()
obj_repr = TruncatedRepr(base_repr)


def initialize(config: Mapping):
    """
    Initializes the module with common configuration settings.
    """
    global EVENT_FIELD_REPR, STACK_FILTER
    EVENT_FIELD_REPR = event_field_repr(config["assess.event_detail"])
    STACK_FILTER = stack_filter(config["assess.stacktraces"])


def event_field_repr(detail: str):
    """
    Returns a function to capture field details in events.
    """
    if detail == "minimal":
        return TruncatedRepr(base_repr)
    elif detail == "full":
        return TruncatedRepr(FullRepr())
    else:
        raise ValueError(f"Unknown event detail: {detail}")


def stack_filter(keep_level: str):
    """
    Returns a function to filter stack frames.
    """
    if keep_level == "ALL":
        return lambda action: False
    elif keep_level == "NONE":
        return lambda action: True
    elif keep_level == "SOME":
        return lambda action: action not in (
            AssessEventAction.CREATION,
            AssessEventAction.TRIGGER,
        )
    elif keep_level == "SINK":
        return lambda action: action != AssessEventAction.TRIGGER
    else:
        raise ValueError(f"Unknown stack level: {keep_level}")


EVENT_FIELD_REPR = event_field_repr("minimal")
STACK_FILTER = stack_filter("ALL")


@dataclass
class Field:
    type: type
    value: str

    def __init__(
        self, obj, typ=None, repr_func: Callable[[object], str] = EVENT_FIELD_REPR.repr
    ):
        self.type = typ or type(obj)
        self.value = repr_func(obj)


class ContrastEvent:
    """
    This class holds the data about an event in the application
    We'll use it to build an event that TeamServer can consume if
    the object to which this event belongs ends in a trigger.
    """

    ATOMIC_ID = 0

    def __init__(
        self,
        node: PolicyNode,
        tagged: Any,
        self_obj: Any | None,
        ret: Any | None,
        args: Sequence[Any],
        kwargs: dict[str, Any],
        parents: list[ContrastEvent],
        possible_key=None,
        source_type: SourceType | None = None,
        source_name=None,
    ):
        self.node = node
        self._init_tagged(tagged, possible_key, args, kwargs)
        self.source_type = source_type
        self.source_name = source_name
        self.parents = parents or []
        ret = self._update_init_return(node, self_obj, ret)
        self.obj = Field(self_obj) if self_obj is not None else None
        self.ret = Field(ret) if ret is not None else None
        self.args = [Field(arg) for arg in args] if args is not None else []
        self.kwargs = (
            {
                k: Field(f"{k}={obj_repr.repr(v)}", typ=type(v))
                for k, v in kwargs.items()
            }
            if kwargs is not None
            else {}
        )

        # These are needed only at trigger-time but values must be set at init.
        self.time_ns = time.time_ns()
        self.thread = threading.current_thread().ident
        self.event_id = ContrastEvent._atomic_id()

        self.event_action = self.node.build_action()

        self._raw_stack = (
            build_stack()
            if not node.skip_stacktrace and not STACK_FILTER(self.event_action)
            else StackSummary()
        )

        # This must happen at init for stream events to work.
        self._update_method_information()

    def _init_tagged(self, tagged: Any, possible_key=None, args=None, kwargs=None):
        """
        Initialize properties related to tagging.
        - self.tagged
        - self.taint_location
        - self.span_override
        """
        self.taint_location: Location | None = (
            possible_key or self._find_taint_location(args, kwargs)
        )
        if self.taint_location and isinstance(tagged, dict):
            tagged = tagged.get(self.taint_location, None)
        tags = None
        self.tagged_props = get_properties(tagged)
        if self.tagged_props and isinstance(tagged, str):
            self.tagged_props.cleanup_tags()
            tags = [tag for tag in chain.from_iterable(self.tagged_props.tags.values())]
            tagged_repr, truncate_tags = truncate_tainted_string(tagged, tags)
            self.tagged = Field(tagged_repr, repr_func=base_repr.repr)
            self.tags_override = truncate_tags
        else:
            self.tagged = Field(tagged, repr_func=TruncatedRepr(FullRepr()).repr)
            self.tags_override = (
                None
                if len(self.tagged.value) == len_or_zero(tagged)
                else (Tag(len(self.tagged.value), 0),)
            )

    def _update_init_return(self, node, obj, ret):
        """
        For purposes of pretty reporting in Teamserver, we will say the
        `__init__` instance method return the `self` object (the object getting
        instantiated) instead of `None`, even though `None` is the return value of
        `__init__` methods.

        This will not apply if the node is not a class (it's possible someone
        creates a module level function called `__init__` or if the return
        value is already populated (for safety).
        """
        if node.method_name == INIT and node.class_name is not None:
            ret = ret or obj
        return ret

    @property
    def parent_ids(self):
        return [parent.event_id for parent in self.parents]

    @classmethod
    def _atomic_id(cls):
        ret = cls.ATOMIC_ID
        cls.ATOMIC_ID += 1
        return ret

    def _find_taint_location(self, args, kwargs) -> Location | None:
        """
        Find the location of the tagged value in the event. This is used to determine
        where the tagged value is in the event, so that we can mark it up in TeamServer.
        """
        if len(self.node.targets) == 1:
            return self.node.targets[0]
        for loc in chain(self.node.targets, self.node.sources):
            if loc in (OBJECT, RETURN):
                return loc
            if isinstance(loc, int) and loc < len(args):
                return loc
            if isinstance(loc, str) and loc in kwargs:
                return loc

        return None

    def _update_method_information(self):
        """
        For nicer reporting, we lie about the tagged value. For example, a call to
        split() returns a list of strings: ["foo", "bar"]. In the properties for "foo",
        the split event shows a return value of only "foo" instead of the whole list.
        """
        if self.taint_location is None:
            # This would be for trigger nodes without source or target. Trigger rule was
            # violated simply by a method being called. We'll save all the information,
            # but nothing will be marked up, as nothing need be tracked.
            return

        if self.taint_location == OBJECT:
            self.obj = self.tagged
            return

        if self.taint_location == RETURN:
            self.ret = self.tagged

    def to_reportable_event(self):
        """
        Convert a ContrastEvent to a AssessEvent.
        """
        event = AssessEvent(
            action=self.event_action,
            args=(
                [self._build_assess_object(arg, i) for i, arg in enumerate(self.args)]
                + [
                    self._build_assess_object(field, key)
                    for key, field in self.kwargs.items()
                ]
            ),
            event_sources=self._build_assess_event_sources(),
            field_name=self.source_name or "",
            object=self._build_assess_object(self.obj, OBJECT),
            object_id=self.event_id,
            parent_object_ids=[AssessParentObject(id=id) for id in self.parent_ids],
            ret=self._build_assess_object(self.ret, RETURN),
            signature=self._build_assess_signature(),
            source=self.node.ts_valid_source,
            target=self.node.ts_valid_target,
            stack=to_assess_stack(self._raw_stack),
            tags=",".join(self.node.tags),
            taint_ranges=[
                AssessTaintRange(tag=tr.tag, range=tr.range)
                for tr in (
                    self.tagged_props.tags_to_ts_obj() if self.tagged_props else []
                )
            ],
            thread=str(self.thread),
            time=self.time_ns // 1_000_000,
            type=self.node.node_type,
        )

        return event

    def build_route_source(self, safe_source_name: str) -> RouteSource:
        """
        Create a new RouteSource
        """
        assert self.source_type is not None
        return RouteSource(self.source_type, safe_source_name)

    def _build_assess_event_sources(self) -> list[AssessEventSource]:
        return (
            [
                AssessEventSource(
                    source_name=self.source_name or "", source_type=self.source_type
                )
            ]
            if self.source_type
            else []
        )

    def _build_assess_object(
        self, field: Field | None, location: Location
    ) -> AssessObject:
        # For now, only the taint_target needs to be officially marked as tracked.
        # This means that agent-tagged strings may not be marked as "tracked" for TS.
        return AssessObject(
            tracked=location == self.taint_location,
            value=base64_encode(field.value) if field else B64_NONE_STRING,
        )

    def _build_assess_signature(self) -> AssessSignature:
        return AssessSignature(
            arg_types=[
                self._type_string(arg) for arg in chain(self.args, self.kwargs.values())
            ],
            class_name=self.node.location.removeprefix("builtins."),
            constructor=self.node.method_name in INITIALIZERS,
            method_name=self.node.method_name,
            return_type=self._type_string(self.ret),
            void_method=False,
            expression_type=None,  # unused by us so far
            operator=None,  # unused by us so far
            signature=None,  # deprecated
            flags=0,  # only for the Java agent
        )

    def _type_string(self, field: Field | None) -> str:
        return field.type.__name__ if field else NONE_STRING

    def history(self: ContrastEvent):
        """
        Return a generator that yields all the events in the history of this event,
        starting with the event itself and then its parents, and so on.

        Events are yielded in depth-first order, so the event itself is yielded first,
        then its last parent, then the last parent of that parent, and so on.

        Events are deduplicated, so if an event is seen more than once, it will not be
        yielded again.
        """
        seen = set()
        queue = [self]
        while queue:
            event = queue.pop()
            seen.add(event)
            yield event
            queue.extend(event for event in event.parents if event not in seen)

    def dot_repr(self: ContrastEvent) -> str:
        """
        Returns a DOT graph representation of self's history.
        """
        return str(DotGraph(self))


@dataclass
class DotGraph:
    event: ContrastEvent
    normalize: bool = False
    _seen_events: dict[ContrastEvent, int] = field(default_factory=dict, init=False)

    def __str__(self):
        dot_lines = [
            "digraph {",
            "   node [shape=plain];",
        ]
        for event in self.event.history():
            dot_lines.append(self._node(event))
            dot_lines.extend(self._edge(parent, event) for parent in event.parents)
        dot_lines.append("}")
        return "\n".join(dot_lines)

    def _node(self, event: ContrastEvent) -> str:
        return f"""{self._event_id(event)} [label=<
        <table cellborder="0" cellspacing="10" style="rounded" {self._tooltip(event)}>
        <tr><td align="text" cellpadding="1"><b>{event.node}</b></td></tr>
        <tr><td><table cellborder="1" cellspacing="0" cellpadding="15">
            <tr><td align="text">data</td><td align="text"><font face="Monospace">{html_escape(event.tagged.value)}</font></td></tr>
            {"".join(f"<tr><td align='text'>{tag}</td><td align='text'><font face='Monospace'>{self._tagrng_markup(rngs, len(event.tagged.value), untagged_marker='&nbsp;')}</font></td></tr>" for tag, rngs in (event.tagged_props.tags.items() if event.tagged_props else []))}
        </table></td></tr>
        </table>>];"""

    def _edge(self, parent: ContrastEvent, child: ContrastEvent) -> str:
        return f"{self._event_id(parent)} -> {self._event_id(child)};"

    def _event_id(self, event: ContrastEvent) -> int:
        if self.normalize:
            if event not in self._seen_events:
                self._seen_events[event] = len(self._seen_events) + 1
            return self._seen_events[event]

        return event.event_id

    def _tooltip(self, event: ContrastEvent) -> str:
        if self.normalize:
            return ""

        # intentionally avoid using clean_stack because it formats filenames
        # replacing / with ., which makes them unusable as links.
        frame = next(
            (frame for frame in reversed(event._raw_stack) if acceptable_frame(frame)),
            None,
        )

        return (
            f'tooltip="{html_escape(frame.line)}" href="file://{frame.filename}"'
            if frame
            else ""
        )

    def _tagrng_markup(self, rngs, length, tag_marker="*", untagged_marker=" ") -> str:
        return "".join(
            tag_marker if any(rng.covers(i) for rng in rngs) else untagged_marker
            for i in range(length)
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from dataclasses import dataclass
from functools import cached_property
from collections import namedtuple

from contrast.agent import scope

from contrast_vendor.importlib_metadata import metadata, PackageMetadata
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

Version = namedtuple("Version", ["major", "minor", "patch"])
DEFAULT_FRAMEWORK = "wsgi"
SUPPORTED_SERVERS = ["uwsgi", "gunicorn", "uvicorn", "mod_wsgi"]
DEFAULT_SERVER = "Unknown"


class DiscoverablePackage:
    def __init__(self, framework_name, default_package):
        self._name = ""
        self.version = None
        self.packages = framework_name if framework_name else DEFAULT_FRAMEWORK
        self.default_package = default_package
        self.set_info()

    @property
    def name(self):
        return self._name.capitalize()

    @property
    def name_lower(self):
        return self._name.lower()

    @property
    def full_version(self):
        return f"{self.version.major}.{self.version.minor}.{self.version.patch}"

    def set_info(self):
        raise NotImplementedError("Must implement set_info")

    def discover_server(self) -> PackageMetadata | None:
        """
        Except in the agent's own testing environment, the assumption here is
        that all environments using the agent will have only
        one supported framework or server.
        """
        with scope.contrast_scope():
            for framework_name in self.packages:
                try:
                    return metadata(framework_name)
                except Exception:  # noqa: PERF203
                    continue

        return None

    def __repr__(self):
        return f"{self.name} {self.full_version}"


class _ServerTypeFramework(DiscoverablePackage):
    """
    Unless you're reporting server type on startup, use `Framework` instead.

    This is only kept because it is used as the server type sent on startup, which the backend includes in
    its server identifier. It's not displayed anywhere in the UI, but changing this value would create new
    servers, leading to loss of settings, more license usage, and more pressure on the backend.
    """

    def __init__(self, framework_name):
        super().__init__(framework_name, DEFAULT_FRAMEWORK)

    def discover_framework(self, framework_name: str) -> PackageMetadata | None:
        """
        Except in the agent's own testing environment, the assumption here is
        that all environments using the agent will have only
        one supported framework or server.
        """
        with scope.contrast_scope():
            try:
                return metadata(framework_name)
            except Exception:
                return None

    def set_info(self):
        framework_metadata = (
            self.discover_framework(self.packages) if self.packages != "wsgi" else None
        )
        if framework_metadata:
            version = framework_metadata.get("Version").split(".")
            patch = version[2] if len(version) > 2 else "0"
            self.version = Version(major=version[0], minor=version[1], patch=patch)
            self._name = framework_metadata.get("Name")
        else:
            self._name = "wsgi"
            self.version = Version(major="0", minor="0", patch="0")


class Server(DiscoverablePackage):
    """
    A class to store information about the current web server used in an application
    """

    def __init__(self):
        super().__init__(SUPPORTED_SERVERS, DEFAULT_SERVER)

    def set_info(self):
        server_metadata = self.discover_server()
        if server_metadata:
            version = server_metadata.get("Version").split(".")
            patch = version[2] if len(version) > 2 else "0"
            self.version = Version(major=version[0], minor=version[1], patch=patch)
            self._name = server_metadata.get("Name")
        else:
            logger.debug(
                "Did not find the current %s. Assuming it's %s.",
                self.__class__.__name__,
                self.default_package,
            )
            self._name = self.default_package
            self.version = Version(major="0", minor="0", patch="0")


@dataclass
class Framework:
    name: str

    def __post_init__(self):
        self.name = self.name.capitalize()

    @cached_property
    def version(self) -> Version | None:
        """
        Get the version of the package by its distribution name.
        """
        if self == UNKNOWN_FRAMEWORK:
            return None
        if dist := metadata(self.name):
            version_parts = dist.get("Version", "0.0.0").split(".")
            return Version(
                major=version_parts[0],
                minor=version_parts[1],
                patch=version_parts[2] if len(version_parts) > 2 else "0",
            )
        return None

    def __str__(self):
        return (
            f"{self.name} {self.version.major}.{self.version.minor}.{self.version.patch}"
            if self.version
            else self.name
        )


UNKNOWN_FRAMEWORK = Framework("UNKNOWN")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import io

from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.utils import (
    copy_events,
    copy_tags_in_span,
    copy_tags_to_offset,
    get_properties,
    set_properties,
)


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def _get_class_name(obj):
    return obj.__class__.__name__


def _get_stream_len(stream, curpos):
    stream.seek(0, io.SEEK_END)
    size = stream.tell()
    stream.seek(curpos)
    return size


def _propagate_stream_read(node, preshift, target, tag_span, ret):
    """Copies tags and creates event for stream read operations"""
    source_properties = get_properties(preshift.obj)
    target_properties = copy_tags_in_span(target, source_properties, tag_span)
    if target_properties is None:
        return

    copy_events(target_properties, source_properties)
    parents = [source_properties.event] if source_properties.event else []
    target_properties.build_event(
        node, target, preshift.obj, ret, preshift.args, preshift.kwargs, parents
    )


def propagate_stream_read(method_name, preshift, target, ret):
    class_name = _get_class_name(preshift.obj)
    node = PropagationNode(
        "io", class_name, True, method_name, "OBJ", "RETURN", "STREAM_READ"
    )

    curpos = preshift.obj.tell()
    startpos = curpos - len(target)
    tag_span = (startpos, curpos)

    _propagate_stream_read(node, preshift, target, tag_span, ret)


def propagate_stream_readlines(method_name, preshift, target, ret):
    # Nothing to do if there are no results
    if len(target) < 1:
        return

    class_name = _get_class_name(preshift.obj)
    node = PropagationNode(
        "io", class_name, True, method_name, "OBJ", "RETURN", "STREAM_READ"
    )

    curpos = preshift.obj.tell()
    startpos = curpos - sum(len(x) for x in target)

    # Readlines returns a list of lines (including newlines)
    for line in target:
        tag_span = (startpos, startpos + len(line))
        _propagate_stream_read(node, preshift, line, tag_span, ret)
        startpos += len(line)


def propagate_stream_getvalue(method_name, preshift, target, ret):
    class_name = _get_class_name(preshift.obj)
    node = PropagationNode(
        "io", class_name, True, method_name, "OBJ", "RETURN", "STREAM_GETVALUE"
    )

    tag_span = (0, len(target))

    _propagate_stream_read(node, preshift, target, tag_span, ret)


def _copy_old_props_from_stream(stream, startpos, curpos):
    if not stream.cs__tracked:
        return None

    old_props = get_properties(stream)
    new_props = old_props.copy()
    # Tags will be updated below
    new_props.tags = {}

    # Copy all tags from the stream prior to the newly written data
    prefix_tags = old_props.tags_at_range(AdjustedSpan(0, startpos))
    copy_tags_to_offset(new_props, prefix_tags, 0)

    # Copy all tags from the stream after the newly written data
    streamlen = _get_stream_len(stream, curpos)
    postfix_tags = old_props.tags_at_range(AdjustedSpan(curpos, streamlen))
    copy_tags_to_offset(new_props, postfix_tags, curpos)

    return new_props


def propagate_stream_write(method_name, preshift, target, ret):
    # This would indicate an error made by the caller, so we can just quit
    if len(preshift.args) < 1:
        return

    # If the stream is a source, no action required (for now)
    # We don't really expect this to occur, and if it does occur frequently it means
    # the stream involved should potentially not be marked as a source after all.
    # The current behavior means that any data written to a source stream will
    # be tagged as UNTRUSTED when it gets read out, regardless of whether it
    # was tracked or not to begin with.
    if preshift.obj.cs__source:
        logger.debug("Data written to source stream; no propagation occurred")
        return

    # Think of write as a special case of writelines
    lines = preshift.args[0] if method_name == "writelines" else [preshift.args[0]]
    props = [get_properties(line) for line in lines]
    # Short circuit here if none of the data, new or old, is tracked
    if not any(props) and not preshift.obj.cs__tracked:
        return

    # Get the current position of the stream cursor (after the write has finished)
    curpos = preshift.obj.tell()
    # Compute the position of the stream cursor before the write occurred
    startpos = curpos - sum([len(line) for line in lines])

    # Copy any tags from the original stream outside of the span of newly written data
    new_props = _copy_old_props_from_stream(preshift.obj, startpos, curpos)

    parents = []

    # TODO: PYT-911 this is not quite right. We make it look like the most recent event on
    # the original string contributes to the `write` event, but that's not really
    # accurate. Instead, the write event and the stream event should contribute to any
    # subsequent `read` events. But this requires an additional level of bookkeeping.
    # We'll pass for now since it only affects trace reports, and not accuracy.
    if new_props is not None and new_props.event:
        parents.append(new_props.event)

    # Iterate over all newly written lines and copy tags if the line is tracked
    for line, source_props in zip(lines, props):
        if source_props is not None:
            # This means that the original stream was not tracked at all
            if new_props is None:
                new_props = source_props.copy(tag_offset=startpos)
            else:
                copy_tags_to_offset(new_props, source_props.tags, startpos)
                copy_events(new_props, source_props)

            if source_props.event:
                parents.append(source_props.event)

        startpos += len(line)

    # Update with new properties on stream object
    preshift.obj.cs__tracked = True
    set_properties(preshift.obj, new_props)

    # Add event
    class_name = _get_class_name(preshift.obj)
    node = PropagationNode(
        "io", class_name, True, method_name, "ARG_0", "OBJ", "STREAM_WRITE"
    )
    new_props.build_event(
        node, target, preshift.obj, ret, preshift.args, preshift.kwargs, parents
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode

re_propagators = [
    CompositeNode(
        {
            "module": "re",
            "target": "RETURN",
            # May be overridden by some individual nodes
            "policy_patch": False,
        },
        [
            {
                "method_name": "split",
                "source": "ARG_1,KWARG:string",
                "action": "REGEX_SPLIT",
                "policy_patch": True,
            },
            {
                "class_name": "Pattern",
                "method_name": "split",
                "source": "ARG_0,KWARG:string",
                "action": "REGEX_SPLIT",
            },
            {
                "method_name": ["sub", "subn"],
                "class_name": [None, "Pattern"],
                "source": "ARG_0,KWARG:string",
                "action": "NONE",
            },
            {
                # re.findall is a simple wrapper around re.Pattern.findall so this single
                # propagation node accounts for both
                "class_name": "Pattern",
                "method_name": "findall",
                "source": "ARG_0,KWARG:string",
                "action": "REGEX_FINDALL",
            },
            {
                "method_name": "escape",
                "source": "ARG_0,KWARG:pattern",
                "action": "SPLAT",
                "policy_patch": True,
            },
            {
                "class_name": "Match",
                "method_name": ["group", "groups", "groupdict"],
                "source": "OBJ",
                "action": "NONE",
            },
        ],
    )
]

register_propagation_nodes(re_propagators)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from contrast import AGENT_CURR_WORKING_DIR
from contrast.agent.agent_lib.input_tracing import InputAnalysisResult
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.protect.rule.base_rule import BaseRule
from collections.abc import Iterable

from contrast.api.sample import Sample
from contrast.utils.decorators import fail_quietly

PARENT_CHECK = ".."
SLASH = "/"
SAFE_PATHS = ["tmp", "public", "docs", "static", "template", "templates"]
WRITE_OPTIONS = ["w", "a"]


class PathTraversal(BaseRule):
    RULE_NAME = "path-traversal"

    def build_sample(
        self,
        evaluation: InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        path = candidate_string
        sample = self.build_base_sample(evaluation)
        if path is not None:
            sample.details["path"] = path
        return sample

    @fail_quietly(
        "Failed to run path traversal skip_protect_analysis", return_value=False
    )
    def skip_protect_analysis(
        self, user_input: object, args: tuple, kwargs: dict[str, object]
    ) -> bool:
        write = possible_write(args, kwargs)
        if write:
            # any write is a risk so we should not skip analysis
            return False

        return not actionable_path(user_input)

    def infilter_kwargs(self, user_input: str, patch_policy: PatchLocationPolicy):
        return dict(method=patch_policy.method_name)


def possible_write(args: tuple, kwargs: dict[str, object]):
    return _possible_write_kwargs(kwargs) or _possible_write_args(args)


def _possible_write_kwargs(kwargs: dict[str, object]) -> bool:
    mode = kwargs.get("mode", "")

    if not isinstance(mode, str):
        return False

    return bool(mode and any([x in mode for x in WRITE_OPTIONS]))


def _possible_write_args(args: tuple) -> bool:
    if not isinstance(args, Iterable):
        return False

    return (
        len(args) > 1
        and args[1] is not None
        and isinstance(args[1], Iterable)
        and any([x in args[1] for x in WRITE_OPTIONS])
    )


def actionable_path(path: object) -> bool:
    if not path or not isinstance(path, str):
        return False

    # moving up directory structure is a risk and hence actionable
    if path.find(PARENT_CHECK) > 1:
        return True

    if "/contrast/" in path or "/site-packages/" in path:
        return False

    if path.startswith(SLASH):
        for prefix in _safer_abs_paths():
            if path.startswith(prefix):
                return False
    else:
        for prefix in SAFE_PATHS:
            if path.startswith(prefix):
                return False

    return True


def _safer_abs_paths() -> list[str]:
    return (
        [f"{AGENT_CURR_WORKING_DIR}/{item}" for item in SAFE_PATHS]
        if AGENT_CURR_WORKING_DIR
        else []
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Using logger.info/debug/someOtherLevel() is not supported in this module. In order to get the correct
frame info, we must skip over functions called in this module and in vendored structlog. If logging is attempted,
incorrect frame info will be displayed on the log message if used in this file.

Use print(...) instead
"""

from __future__ import annotations

import asyncio
from contextlib import suppress
import os
import pathlib
import io
from socket import gethostname
import sys
from dataclasses import dataclass
from typing import TextIO, cast

from contrast.agent import request_state
from contrast.utils.namespace import Namespace
from contrast_vendor import structlog
from contrast_vendor.filelock import FileLock, Timeout


class module(Namespace):
    file_handle: TextIO | None = None


@dataclass
class RotationConfig:
    backup_count: int
    max_bytes: int


class RotatingFile:
    """
    A file handler that rotates the log file when it reaches a certain size.

    Writes to instances of this class are not thread-safe. If calling write() from multiple threads,
    you must ensure that the writes are synchronized externally.
    """

    def __init__(self, file: TextIO, config: RotationConfig):
        if config.max_bytes <= 0:
            raise ValueError("max_bytes must be greater than 0")
        self._max_bytes = config.max_bytes
        if config.backup_count <= 0:
            raise ValueError("backup_count must be greater than 0")
        self._backup_count = config.backup_count
        self._file = file
        self._filename = file.name
        self._lock = FileLock(f".{self._filename}.lock")

    def write(self, message: str) -> int:
        """Write a message to the log file, performing rotation if needed."""
        self._rotate_if_needed(len(message))

        return self._file.write(message)

    # NOTE: flush() and close() are intentionally hardcoding the implementation
    # of __getattr__. This is critical for flush(), because structlog's WriteLogger
    # binds the flush method function to an internal attribute. If we don't introduce
    # this layer of indirection, then when we rotate the _file, structlog will attempt
    # to call flush() on the old file handle, which has already been closed.

    def flush(self):
        """Flush the log file."""
        self._file.flush()

    def close(self):
        """Close the log file."""
        self._file.close()

    def _reopen(self):
        """Reopen the log file."""
        self._file.close()
        self._file = open(self._filename, "a", encoding="utf-8")  # noqa: SIM115

    def _should_rollover(self, msg_size: int):
        """Check if the log file should be rotated."""
        file_size = self._file.tell()
        if file_size == 0:
            # If the file is empty, we either need to write to it or
            # discard the message. We can't rollover, because the new
            # file will be empty and we'd enter an infinitely loop.
            return False
        return file_size + msg_size > self._max_bytes

    def _rotate_if_needed(self, msg_size: int):
        """Check if the log file should be rotated and perform the rotation if needed."""
        # Refer to https://en.wikipedia.org/wiki/Double-checked_locking for background
        # on the pattern used here (and ways it can be broken).
        if not self._should_rollover(msg_size):
            return

        while True:
            with suppress(Timeout), self._lock.acquire(timeout=1):
                self._reopen()
                if self._should_rollover(msg_size):
                    self._do_rollover()
                return

    def _do_rollover(self):
        """Perform log file rotation."""
        # This implementation follows the Java and .NET agents decisions.
        # It's a little unconventional, but we want to keep the same behavior across languages
        # so that it's easier for our support team to troubleshoot issues from the logs.

        # Check if the highest backup slot is filled
        highest_backup = f"{self._filename}.{self._backup_count}"
        if not os.path.exists(highest_backup):
            # Not all slots are filled yet - find next available slot
            next_slot = 1
            while (
                os.path.exists(f"{self._filename}.{next_slot}")
                and next_slot < self._backup_count
            ):
                next_slot += 1

            os.rename(self._filename, f"{self._filename}.{next_slot}")
        else:
            # All slots filled - need to shift files
            # contrast.log.1 is the oldest, contrast.log.N is newest

            # Delete the oldest backup
            os.remove(f"{self._filename}.1")

            # Shift files down by one position (2→1, 3→2, etc.)
            for i in range(1, self._backup_count):
                src = f"{self._filename}.{i + 1}"
                dst = f"{self._filename}.{i}"
                if os.path.exists(src):
                    os.rename(src, dst)

            os.rename(self._filename, highest_backup)

        # Reopen the file
        self._reopen()

    def __getattr__(self, name: str):
        return getattr(self._file, name)


def add_hostname(logger, method_name, event_dict):
    event_dict["hostname"] = gethostname()
    return event_dict


def add_request_id(logger, method_name, event_dict):
    event_dict["request_id"] = request_state.get_request_id()
    return event_dict


def rename_key(old_name, new_name):
    def key_renamer(logger, method_name, event_dict):
        value = event_dict.get(old_name)
        if value and not event_dict.get(new_name):
            event_dict[new_name] = value
            del event_dict[old_name]

        return event_dict

    return key_renamer


def add_progname(logger, method_name, event_dict):
    """
    progname is the name of the process the agents uses in logs.
    The default value is Contrast Agent. progname will be used
    as the name of the logger as seen in the logs.
    """
    event_dict["name"] = "Contrast Agent"

    return event_dict


def add_asyncio_info(logger, method_name, event_dict):
    try:
        if (current_task := asyncio.current_task()) is not None:
            # If no name has been explicitly assigned to the Task, the default asyncio Task implementation
            # generates a default name during instantiation.
            event_dict["asyncio_task_name"] = current_task.get_name()
            current_coro = current_task.get_coro()
            event_dict["asyncio_coro_name"] = current_coro.__qualname__
            event_dict["asyncio_task_id"] = id(current_task)
    except Exception:
        # This can happen when there is no running event loop
        pass

    return event_dict


def _log_level_to_int(level: str) -> int:
    level_to_int = {
        "DEBUG": 10,
        "TRACE": 10,
        "INFO": 20,
        "WARNING": 30,
        "ERROR": 40,
        "CRITICAL": 50,
    }

    level = level.upper()
    if level == "TRACE":
        sys.stderr.write(
            "Contrast Python Agent: TRACE logging is equivalent to DEBUG\n"
        )

    return level_to_int.get(level, level_to_int["INFO"])


def _close_handler():
    handle = module.file_handle
    if handle not in (sys.stderr, sys.stdout, None):
        module.file_handle = None
        handle.close()


def _set_handler(filename: TextIO | str, rotation_config: RotationConfig | None = None):
    if isinstance(filename, str):
        try:
            path = pathlib.Path(filename).parent.resolve()
            os.makedirs(path, exist_ok=True)
            module.file_handle = open(filename, "a", encoding="utf-8")  # noqa: SIM115
            if rotation_config:
                module.file_handle = cast(
                    TextIO, RotatingFile(module.file_handle, rotation_config)
                )

        except Exception as e:
            sys.stderr.write(f"Failed to create log file {filename} - {e}\n")
            module.file_handle = sys.stderr
    elif isinstance(filename, io.TextIOBase):
        module.file_handle = filename
    else:
        raise TypeError(
            "log_file must be a string path to a file or an open TextIOBase object",
            type(filename).__name__,
        )

    return module.file_handle


def init_structlog(
    log_level_name: str,
    log_file: TextIO | str,
    *,
    # NOTE: We should only enable logger caching if its configuration is finalized. If
    # it's possible for the logging config to change in the future, do not cache it.
    cache_logger: bool = False,
    rotation_config: RotationConfig | None = None,
) -> None:
    """
    Initial configuration for structlog. This can still be modified by subsequent calls
    to structlog.configure.
    """
    log_level = _log_level_to_int(log_level_name)

    _close_handler()
    file_handle = _set_handler(log_file, rotation_config)

    structlog.configure(
        processors=[
            structlog.processors.MaybeTimeStamper(fmt="iso", key="time"),
            structlog.processors.add_log_level,
            rename_key("event", "msg"),
            structlog.processors.CallsiteParameterAdder(
                [
                    structlog.processors.CallsiteParameter.PROCESS,
                    structlog.processors.CallsiteParameter.FILENAME,
                    structlog.processors.CallsiteParameter.FUNC_NAME,
                    structlog.processors.CallsiteParameter.LINENO,
                    structlog.processors.CallsiteParameter.THREAD,
                    structlog.processors.CallsiteParameter.THREAD_NAME,
                ],
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ],
            ),
            rename_key("process", "pid"),
            add_request_id,
            add_asyncio_info,
            add_hostname,
            add_progname,
            structlog.processors.format_exc_info,
            structlog.processors.StackInfoRenderer(
                additional_ignores=[
                    "contrast_vendor.structlog",
                    "contrast.utils.decorators",
                ]
            ),
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer(),
        ],
        logger_factory=structlog.WriteLoggerFactory(file_handle),
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        cache_logger_on_first_use=cache_logger,
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_header_only_rule import (
    BaseHeaderOnlyRule,
)


class CspHeaderMissingRule(BaseHeaderOnlyRule):
    """
    Rule is violated if all the following headers:

    - Content-Security-Policy
    - X-Content-Security-Policy
    - X-Webkit-CSP

    are missing.
    """

    @property
    def name(self):
        return "csp-header-missing"

    @property
    def header_key(self):
        """We care about multiple headers for this rule, see `header_keys`"""

    @property
    def header_keys(self):
        return ("Content-Security-Policy", "X-Content-Security-Policy", "X-Webkit-CSP")

    @property
    def good_values(self):
        """We don't care about the values for this rule, only header(s) presence"""

    def is_header_violated(self, headers):
        for key in self.header_keys:
            value = headers.get(key)
            if value:
                # As long as one of the headers is present,  rule is not violated.
                return False, None

        # Finding gets no properties if header(s) is missing
        return True, {}



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.policy_node import PolicyNode


class DeadZoneNode(PolicyNode):
    def __init__(
        self,
        module,
        class_name,
        method_name,
        instance_method=True,
        config_option=None,
        policy_patch=True,
    ):
        super().__init__(
            module=module,
            class_name=class_name,
            method_name=method_name,
            instance_method=instance_method,
            policy_patch=policy_patch,
        )
        self.config_option = config_option



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode


webob_sources = [
    CompositeNode(
        {
            "module": "webob.multidict",
            "class_name": "MultiDict",
        },
        [
            {
                "method_name": ["__getitem__", "getall", "mixed", "dict_of_lists"],
                "node_type": "PARAMETER",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    CompositeNode(
        {
            "module": "webob.request",
            "class_name": "BaseRequest",
        },
        [
            {
                "method_name": "query_string",
                "node_type": "QUERYSTRING",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
            {
                "method_name": "host_url",
                "node_type": "URI",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
        ],
    ),
    {
        "module": "webob.headers",
        "class_name": "EnvironHeaders",
        "method_name": "__getitem__",
        "node_type": "HEADER",
        "tags": ["NO_NEWLINES"],
    },
    {
        "module": "webob.cookies",
        "class_name": "RequestCookies",
        "method_name": "__getitem__",
        "node_type": "COOKIE",
        "tags": ["NO_NEWLINES"],
    },
]


register_source_nodes(webob_sources)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor.webob.headers import ResponseHeaders
from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseResponseWrapper,
)


class AioHttpResponseWrapper(BaseResponseWrapper):
    def __init__(self, response):
        self._response = response
        self._streaming_cache = None

    @property
    def body(self):
        return self._response.body

    @property
    def headers(self):
        return ResponseHeaders(self._response.headers)

    @property
    def status_code(self):
        return self._response.status



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from functools import cached_property
from typing import TYPE_CHECKING

from contrast.agent import scope

from contrast.agent.agent_lib.input_tracing import InputAnalysisResult
from contrast.agent.exclusions import Exclusions
from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseResponseWrapper,
)
from contrast.agent.request import Request
from contrast.agent.settings import Settings
from contrast.agent.inventory.library_reader import LibraryObservations
from contrast_fireball import AssessFinding, ObservedRoute
from contrast.api.attack import Attack
from contrast.reporting.fireball import ObservabilityTrace
from contrast_vendor import structlog as logging
from contrast.utils.digest_utils import Digest
from contrast.utils.decorators import fail_quietly
from contrast.reporting.request_masker import RequestMasker

if TYPE_CHECKING:
    from contrast.agent.policy.registry_v2 import EventHandler

logger = logging.getLogger("contrast")


class RequestContext:
    @scope.contrast_scope()
    def __init__(
        self,
        environ,
        exclusions: Exclusions | None = None,
        request_data_masker: RequestMasker | None = None,
        event_handlers: dict[str, list[EventHandler]] | None = None,
        *,
        assess_enabled: bool = False,
        observe_enabled: bool = False,
        protect_enabled: bool = False,
    ):
        self.request = Request(environ)
        self.response = None
        self.response_exception: Exception | None = None

        # This contains any Observed Library Usage seen during the life of this request.
        self.observed_libraries = LibraryObservations([])

        self.exclusions = exclusions
        self.event_handlers: dict[str, list[EventHandler]] = event_handlers or {}

        self.observability_trace: ObservabilityTrace | None = None

        self.request_data_masker = None

        if request_data_masker:
            self.request_data_masker = RequestMasker.new_request_masker(
                request_data_masker.mask_rules
            )

        # This contains input exclusions that should always apply regardless of rule on this request
        self.input_exclusions = None

        # This contains exclusions that only apply at trigger time based on the source event
        #
        # Storing this attribute here makes typing difficult.
        # RequestContext now depends on InputExclusion which depends on RequestContext.
        self.input_exclusions_trigger_time = []

        self.excluded_assess_rules = []
        self.excluded_protect_rules = []

        # For protect: store attacks made during a request to report at the end
        self.attacks: list[Attack] = []

        # For assess: store findings made during a request to report at the end
        self.findings: list[AssessFinding] = []

        self.user_input_analysis: list[InputAnalysisResult] = []
        self.response = None
        self.do_not_track = False

        self.signature: str | None = None
        self.path_template: str | None = None
        """
        Framework-specific representation of the URL structure for this request. This
        must exactly match the corresponding discovered route's URL. It also must not
        contain any user-specific data (ie never requires masking).

        This is not a normalized URL. Using Contrast's normalization algorithm on the
        concrete path for this request will not produce a correct path template. If a
        path template cannot be obtained from the framework, do not set this value.

        Example (flask): "/users/<user_id>/posts/<int:post_id>"
        """

        self.observed_route = ObservedRoute("", "", "", [])

        self.assess_enabled = assess_enabled
        self.observe_enabled = observe_enabled
        self.protect_enabled = protect_enabled

        self.source_count = 0
        self.propagation_count = 0
        self.max_sources_logged = False
        self.max_propagators_logged = False

    @property
    def disabled(self) -> bool:
        return (
            self.assess_enabled is False
            and self.protect_enabled is False
            and self.observe_enabled is False
        )

    @property
    def observed_library_usage(self) -> list[LibraryObservations]:
        """
        Returns a list of library observations for libraries imported within the current request.
        """
        return self.observed_libraries.to_list()

    @cached_property
    @fail_quietly("Unable to compute request hash")
    def hash(self) -> str:
        """
        Generates a CRC32 checksum for the request based on the request method,
        normalized uri and content length.
        """
        hasher = Digest()

        hasher.update(self.request.method + self.request.get_normalized_uri())
        return hasher.finish()

    def stop_source_creation(self, source_type, source_name):
        """
        Compare `source_count` to `max_context_source_events` config option
        :return: true if source_count within this request is equal to or
                 greater than configured threshold for source creation
        """
        if self.max_sources_logged:
            return True

        if self.exclusions and self.exclusions.evaluate_input_exclusions(
            self, source_type, source_name
        ):
            return True

        threshold_reached = self.source_count >= Settings().max_sources
        if threshold_reached:
            logger.warning(
                "Will not create more sources in this request. %s sources reached",
                self.source_count,
            )
            self.max_sources_logged = True

        return threshold_reached

    @property
    def stop_propagation(self):
        """
        Compare `propagation_count` to `max_propagation_events` config option
        :return: true if propagation_count within this request is equal to or
                 greater than configured threshold for propagation
        """
        if self.max_propagators_logged:
            return True

        threshold_reached = self.propagation_count >= Settings().max_propagation
        if threshold_reached:
            logger.warning(
                "Will not propagate any more in this request. %s propagations reached.",
                self.propagation_count,
            )
            self.max_propagators_logged = True

        return threshold_reached

    @property
    def propagate_assess(self):
        # TODO: PYT-644 move this property of out this class?
        return self.assess_enabled and not scope.in_contrast_or_propagation_scope()

    def source_created(self):
        """
        Increase the running count of sources created in this request
        """
        self.source_count += 1

    def propagated(self):
        """
        Increase the running count of propagations created in this request
        """
        self.propagation_count += 1

    def extract_response(self, response: BaseResponseWrapper):
        """
        Fully generate / iterate over the response body and store a reference to the
        response object on this request context.

        This function must run as though it were running in the application directly; it
        must not run in scope, and any errors during body iteration must be raised
        externally (not caught by Contrast internal machinery).

        Some applications generate response bodies lazily, meaning content is produced
        at response body iteration time. This is particularly common in apps using html
        template rendering. It's important that body iteration does not occur in scope
        so that dataflow will occur normally during content generation.
        """
        assert not scope.in_contrast_scope()
        _ = response.body
        self.response = response

    def evaluate_exclusions(self) -> bool:
        """
        Check if the request is excluded based on the URL or input exclusions.

        :return: True if the request is excluded, False otherwise
        """
        if self.exclusions:
            path = self.request.path_info
            self.exclusions.evaluate_url_exclusions(self, path)
            if self.disabled:
                # Stop analyzing this endpoint since the URL exclusion applies
                return True

            self.exclusions.set_input_exclusions_by_url(self, path)

        return False



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Generator

import uuid
import contextlib
from contextvars import ContextVar

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


_request_id: ContextVar[object | None] = ContextVar("contrast_request_id", default=None)


@contextlib.contextmanager
def request_id_context() -> Generator[None, None, None]:
    """
    Context manager for request_id. Produces a uuid contextvar for uniquely identifying
    a request across contrast logs. This should be set as early as possible during
    request processing.
    """
    token = _request_id.set(uuid.uuid4().hex)
    try:
        yield
    finally:
        _request_id.reset(token)


def get_request_id() -> object | None:
    return _request_id.get()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

Py_TPFLAGS_HEAPTYPE: int
if sys.version_info >= (3, 10):
    Py_TPFLAGS_IMMUTABLETYPE: int

DEBUG: int

def initialize_logger(agent_logger) -> None: ...
def initialize() -> None: ...
def enable_c_patches() -> None: ...
def disable_c_patches() -> None: ...
def get_tp_version_tag(type: type, /) -> int: ...
def set_attr_on_type(owner: type, name: str, value, /) -> None: ...



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    wrap_and_watermark,
    register_module_patcher,
)


# TODO: PYT-3819 this patch is not currently being applied; uncomment the lines in
# patches/__init__.py to apply it
def build_graphene_schema_init_patch(orig_func, _):
    def schema_init_patch(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)
        assert result is None

        schema = instance

        _ = discover_graphene_routes(schema.graphql_schema.query_type)
        _ = discover_graphene_routes(schema.graphql_schema.mutation_type)

        return result

    return wrap_and_watermark(orig_func, schema_init_patch)


def discover_graphene_routes(query_or_mutation_obj):
    # TODO: PYT-3819 This is just a placeholder to demonstrate some of the introspection
    # operations we can perform on a graphql-core query/mutation object. The actual
    # implementation here will need to change considerably for route discovery to work
    # correctly.
    if not query_or_mutation_obj:
        return {}

    routes = {}
    for name, field in query_or_mutation_obj.fields.items():
        routes[name] = {
            "type": field.type,
            "args": {k: v.type for k, v in field.args.items()},
            # NOTE: `resolve` might not work for special cases like __schema, __type, __typename
            "resolver": field.resolve,
        }

    return routes


def patch_graphene(graphene_module):
    build_and_apply_patch(
        graphene_module.Schema,
        "__init__",
        build_graphene_schema_init_patch,
    )


def register_patches():
    register_module_patcher(patch_graphene, "graphene")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
from abc import ABC, abstractmethod
from contrast.agent import scope
from contrast.agent.settings import Settings
from contrast.reporting.teamserver_messages import ServerActivity
from contrast.reporting.teamserver_messages.application_settings import (
    ApplicationSettings,
)
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import sleep
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")
SERVER_SETTINGS_THREAD_NAME = "ContrastServerSettings"
APPLICATION_SETTINGS_THREAD_NAME = "ContrastApplicationSettings"
MIN_INTERVAL_MS = 10_000


class BaseSettingsThread(ABC, threading.Thread):
    def __init__(self, reporting_client, interval_ms: int, thread_name):
        self.stopped = False
        # Agent should not ping too frequently
        self.settings_interval_ms = max(interval_ms, MIN_INTERVAL_MS)
        self.reporting_client = reporting_client

        super().__init__()
        # A thread must have had __init__ called, but not start, to set daemon
        self.daemon = True
        self.name = thread_name

    def start(self):
        self.stopped = False
        super().start()

    @property
    def settings_interval_sec(self):
        return self.settings_interval_ms / 1000

    def run(self):
        # Ensure the polling thread runs in scope because it is
        # initialized before our thread.start patch is applied.
        with scope.contrast_scope():
            logger.debug("Starting settings polling thread", name=self.name)
            while not self.stopped and Settings().is_agent_config_enabled():
                self.send_settings()
                sleep(self.settings_interval_sec)

    @abstractmethod
    def send_settings(self) -> None: ...


class ServerSettingsThread(BaseSettingsThread):
    def __init__(self, reporting_client):
        super().__init__(
            reporting_client,
            Settings().config.server_settings_poll_interval,
            SERVER_SETTINGS_THREAD_NAME,
        )

    @fail_loudly("Error sending a server settings message")
    def send_settings(self):
        """
        We're currently hitting the Server Activity endpoint to get our server
        features/ settings. This NG style endpoint will eventually be replaced
        with the true v1 Server Settings endpoint, so naming around it reflects
        that. The Server Activity message is always empty, so we're already
        only using it for settings, not reporting. The content of the message
        is for now deprecated java-only features.
        """
        if Settings().config is None:
            return
        msg = ServerActivity()
        response = self.reporting_client.send_message(msg)
        msg.process_response(response, self.reporting_client)


class ApplicationSettingsThread(BaseSettingsThread):
    def __init__(self, reporting_client):
        super().__init__(
            reporting_client,
            Settings().config.application_settings_poll_interval,
            APPLICATION_SETTINGS_THREAD_NAME,
        )

    @fail_loudly("Error sending an application settings message")
    def send_settings(self):
        if Settings().config is None:
            return
        msg = ApplicationSettings()
        response = self.reporting_client.send_message(msg)
        msg.process_response(response, self.reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

import inspect
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum, auto

from contrast.agent.policy import patch_manager
from contrast.utils.namespace import Namespace
from contrast.utils.patch_utils import (
    add_watermark,
    build_and_apply_patch,
    register_module_patcher,
    wrap_and_watermark,
)
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class module(Namespace):
    called_with_middleware: ContextVar = ContextVar(
        "called_with_middleware", default=False
    )
    middlewares = {}


class AppInterfaceType(Enum):
    WSGI = auto()
    ASGI = auto()
    AUTO_DETECT = auto()


@contextmanager
def called_with_middleware():
    """
    We patch `__call__` on the application object and call our middleware from within
    that patch. This contextmanager provides a mechanism to avoid infinite recursion
    when the middleware in turn calls the original application.
    """
    try:
        module.called_with_middleware.set(True)
        yield
    finally:
        module.called_with_middleware.set(False)


def _is_asgi(app_interface: AppInterfaceType, app_call_func) -> bool:
    """
    Determine if we should consider this application to be an ASGI app.

    If an `app_interface` has been explicitly requested, use that. Otherwise,
    automatically determine the interface by inspection.
    """
    return (
        inspect.iscoroutinefunction(app_call_func)
        if app_interface == AppInterfaceType.AUTO_DETECT
        else app_interface == AppInterfaceType.ASGI
    )


def build__init__patch(
    orig_func, patch_policy, framework_name: str, app_interface: AppInterfaceType
):
    """
    Generic patch for __init__ method of various application classes

    This initializes the contrast middleware and saves it to this module
    """
    del patch_policy

    def __init__patch(wrapped, instance, args, kwargs) -> None:
        logger.info(f"Detected application: {framework_name}")

        from contrast.agent.agent_state import automatic_middleware
        from contrast.asgi.middleware import ASGIMiddleware
        from contrast.wsgi.middleware import WSGIMiddleware

        result = wrapped(*args, **kwargs)

        is_asgi = _is_asgi(app_interface, instance.__call__)
        logger.info(f"Application interface: {'ASGI' if is_asgi else 'WSGI'}")

        with automatic_middleware():
            module.middlewares[id(instance)] = (
                ASGIMiddleware if is_asgi else WSGIMiddleware
            )(instance, framework_name=framework_name)

        return result

    return wrap_and_watermark(orig_func, __init__patch)


def build__call__patch(orig_func, patch_policy, app_interface: AppInterfaceType):
    """
    Generic patch for automatically instrumenting a WSGI/ASGI app's __call__ method.

    This patch calls the middleware, which must have been initialized in the `__init__`
    patch. This patch builder is also responsible for determining if we need to supply
    an async __call__ patch (ASGI) or sync (WSGI).

    ----------------------- NOTE -----------------------

    This patch does not make use of wrapt function wrappers. This is because `uvicorn`
    assumes that if the "application object" can be used like this

    ```
    application_object = application_object()
    ```

    without raising a TypeError, then the application object must have actually been an
    application factory. This is extremely unfortunate for us, because it means that the
    signature of our __call__ patch needs to match the original call (we can't accept
    *args, **kwargs and just pass them through).

    We can't even raise a TypeError manually, because in the async case, the TypeError
    wouldn't be raised until the coroutine is awaited. Uvicorn calls the app as though
    it were a synchronous function without actually checking.
    """
    del patch_policy

    async def __call__patch_asgi(self, scope, receive, send):
        if module.called_with_middleware.get():
            return await orig_func(self, scope, receive, send)

        with called_with_middleware():
            return await module.middlewares[id(self)](scope, receive, send)

    def __call__patch_wsgi(self, environ, start_response):
        if module.called_with_middleware.get():
            return orig_func(self, environ, start_response)

        with called_with_middleware():
            return module.middlewares[id(self)](environ, start_response)

    patch = (
        __call__patch_asgi if _is_asgi(app_interface, orig_func) else __call__patch_wsgi
    )
    return add_watermark(patch)


class CommonMiddlewarePatch:
    """
    Class that implements generic application patches for a variety of frameworks

    We expect that this class can potentially be used in any framework where
    the __init__ method of the framework's application class will be used for
    automatic instrumentation.

    :param module_name: Name of the module that owns the application class
    :param framework_name: Name of the framework (defaults to `module_name`)
    :param application_class_name: Name of the application class to be hooked (defaults to capitalized `framework_name`)
    """

    def __init__(
        self,
        module_name: str,
        *,
        application_class_name: str | None = None,
        framework_name: str | None = None,
        app_interface: AppInterfaceType = AppInterfaceType.WSGI,
    ):
        self.module_name = module_name
        self.framework_name = framework_name or module_name
        self.application_class_name = (
            application_class_name or self.framework_name.capitalize()
        )
        self.app_interface = app_interface

    @property
    def __name__(self):
        return f"{__name__.rpartition('.')[0]}.{self.module_name}"

    def register_patches(self):
        """
        Registers post-import hook for the __init__ method of the application class
        """

        def patch_application(module):
            app_class = getattr(module, self.application_class_name)
            build_and_apply_patch(
                app_class,
                "__init__",
                build__init__patch,
                builder_args=(
                    self.framework_name,
                    self.app_interface,
                ),
            )
            build_and_apply_patch(
                app_class,
                "__call__",
                build__call__patch,
                builder_args=(self.app_interface,),
            )

        register_module_patcher(patch_application, self.module_name)

    def reverse_patches(self):
        """
        Reverses patches for the __init__ method of the application class
        """
        patch_manager.reverse_class_patches_by_name(
            self.module_name,
            self.application_class_name,
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.deprecated_middleware import deprecated_wsgi
from contrast.patches.middleware.django import initialize_django

WSGIMiddleware = deprecated_wsgi("django")


class DjangoWSGIMiddleware(WSGIMiddleware):
    def __init__(self, *args, **kwargs):
        initialize_django()
        super().__init__(*args, **kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
from contrast.agent.assess.rules.providers.hardcoded_key import HardcodedKey
from contrast.agent.assess.rules.providers.hardcoded_password import HardcodedPassword
from contrast.utils.patch_utils import get_loaded_modules

PROVIDERS_THREAD_NAME = "ContrastProviders"


def run_providers_in_thread():
    threading.Thread(
        target=_run_providers,
        name=PROVIDERS_THREAD_NAME,
        daemon=True,
    ).start()


def _run_providers():
    """
    Providers are non-dataflow rules that analyze the contents of a module.
    """
    providers = [p for p in [HardcodedKey(), HardcodedPassword()] if not p.disabled]
    if not providers:
        return

    loaded_modules = get_loaded_modules().values()
    for provider in providers:
        for module in loaded_modules:
            provider.analyze(module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.mode import Mode


class Xss(BaseRule):
    """
    Cross Site Scripting Protection rule
    Currently only a prefilter / block at perimeter rule
    """

    RULE_NAME = "reflected-xss"

    def is_prefilter(self):
        return self.enabled

    @property
    def mode(self):
        """
        Always block at perimeter
        """
        mode = self.settings.config.get(self.config_rule_path_mode)

        return Mode.BLOCK_AT_PERIMETER if mode == Mode.BLOCK else mode



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode

os_propagators = [
    CompositeNode(
        {
            "module": "os",
        },
        [
            {
                "method_name": "scandir",
                "source": "ARG_0,KWARG:path",
                "action": "SPLAT",
                "target": "RETURN",
                "policy_patch": False,
            },
        ],
    )
]

register_propagation_nodes(os_propagators)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Contains dependencies of the agent that we have chosen to vendor

Vendoring dependencies can make life easier on both ourselves and our
customers. Some customers have fairly constrained environments which can make
it difficult for them to install our dependencies. By vendoring, we reduce the
burden for them. Vendoring also allows ourselves to use specific, known stable
versions of dependencies without worrying about conflicts in customer
environments. This can be especially useful for common packages that are
expected to already exist in most environments (e.g. `ruamel.yaml`).

The code under this directory should be ignored when linting and when
calculating test coverage since it is technically out of our control. In
general we should not make changes to the code in this directory other than to
introduce a different version of a particular package.

A list of included packages can be found under vendor-requirements.txt.
"""



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

import os
import logging as stdlib_logging

import contrast
from contrast_vendor import structlog as logging
from contrast.agent.validator import ValidationException
from contrast_vendor.webob.request import DisconnectionError

logger = logging.getLogger("contrast")

DEBUG_LEVEL = stdlib_logging.DEBUG


def log_and_report_exception(
    log_message: str | None,
    error: Exception,
    original_func,
    args,
    kwargs,
    log_level="debug",
):
    """
    Log an exception at the given level and report it to telemetry if available.
    """
    TELEMETRY = contrast.TELEMETRY if not contrast.telemetry_disabled() else None

    try:
        full_msg = (
            (log_message or "Exception in Contrast machinery") + ": " + str(error)
        )
        getattr(logger, log_level)(full_msg, exc_info=error, stack_info=True)
        logger.debug("wrapped function args: %s", args)
        logger.debug("wrapped function kwargs: %s", kwargs)

        if TELEMETRY is not None:
            # omit the decorator wrapper function from the stacktrace
            error.__traceback__ = error.__traceback__.tb_next
            TELEMETRY.report_error(
                error=error,
                original_func=original_func,
                message=full_msg,
                # 1 to remove the current _report_exception frame
                skip_frames=1,
            )
    except ValidationException as val_ex:
        logger.debug(f"Cannot report error to Telemetry: {str(val_ex)}")
    except Exception:
        # For complete safety, we're not going to try to log the logging error
        # because sometimes the logging error is actually an error (such as a recursive
        # error) within the logging module itself!
        pass


SILENCED_EXCEPTIONS = (DisconnectionError, ConnectionResetError)


def _fail_safely(log_message, log_level, return_value):
    """
    Decorator that will run the decorated function/method and, if
    an exception is raised, return a safe value and log the error.

    Note that SecurityException will always be re-raised.

    :param log_message: message to log in case of failure
    :param log_level: level to log in case of failure
    :param return_value: safe value to return in case of failure
    :return: original func return or return_value
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except contrast.SecurityException:
                raise
            except Exception as ex:
                TESTING = os.environ.get("CONTRAST_TESTING")
                if isinstance(ex, SILENCED_EXCEPTIONS) and not TESTING:
                    logger.debug("Silenced exception in fail_safely", exc_info=ex)
                    return return_value

                log_and_report_exception(
                    log_message, ex, original_func, args, kwargs, log_level
                )
                if TESTING:
                    logger.warn(
                        "Re-raising exception in fail_safely (CONTRAST_TESTING is set)"
                    )
                    raise

            return return_value

        return run_safely

    return wrap


def fail_loudly(log_message=None, log_level="exception", return_value=None):
    return _fail_safely(log_message, log_level, return_value)


def fail_quietly(log_message="fail_quietly caught an exception", return_value=None):
    """
    Similar to fail_loudly (see above)

    This decorator is intended to handle cases where an exception may occur but won't
    disrupt normal operation of the agent. This decorator should be used to protect
    against external exceptions we can't prevent but still want to handle.

    In these cases, we log an error message and the exception traceback, both at DEBUG
    level.
    """
    return _fail_safely(f"WARNING: {log_message}", "debug", return_value)



import calendar

from datetime import (
    date,
    datetime,
    timedelta,
    tzinfo,
    )

from email.utils import (
    formatdate,
    mktime_tz,
    parsedate_tz,
    )

import time

from contrast_vendor.webob.compat import (
    integer_types,
    long,
    native_,
    text_type,
    )

__all__ = [
    'UTC', 'timedelta_to_seconds',
    'year', 'month', 'week', 'day', 'hour', 'minute', 'second',
    'parse_date', 'serialize_date',
    'parse_date_delta', 'serialize_date_delta',
]

_now = datetime.now # hook point for unit tests

class _UTC(tzinfo):
    def dst(self, dt):
        return timedelta(0)
    def utcoffset(self, dt):
        return timedelta(0)
    def tzname(self, dt):
        return 'UTC'
    def __repr__(self):
        return 'UTC'

UTC = _UTC()



def timedelta_to_seconds(td):
    """
    Converts a timedelta instance to seconds.
    """
    return td.seconds + (td.days*24*60*60)

day = timedelta(days=1)
week = timedelta(weeks=1)
hour = timedelta(hours=1)
minute = timedelta(minutes=1)
second = timedelta(seconds=1)
# Estimate, I know; good enough for expirations
month = timedelta(days=30)
year = timedelta(days=365)


def parse_date(value):
    if not value:
        return None
    try:
        value = native_(value)
    except:
        return None
    t = parsedate_tz(value)
    if t is None:
        # Could not parse
        return None
    if t[-1] is None:
        # No timezone given.  None would mean local time, but we'll force UTC
        t = t[:9] + (0,)
    t = mktime_tz(t)
    return datetime.fromtimestamp(t, UTC)

def serialize_date(dt):
    if isinstance(dt, (bytes, text_type)):
        return native_(dt)
    if isinstance(dt, timedelta):
        dt = _now() + dt
    if isinstance(dt, (datetime, date)):
        dt = dt.timetuple()
    if isinstance(dt, (tuple, time.struct_time)):
        dt = calendar.timegm(dt)
    if not (isinstance(dt, float) or isinstance(dt, integer_types)):
        raise ValueError(
            "You must pass in a datetime, date, time tuple, or integer object, "
            "not %r" % dt)
    return formatdate(dt, usegmt=True)



def parse_date_delta(value):
    """
    like parse_date, but also handle delta seconds
    """
    if not value:
        return None
    try:
        value = int(value)
    except ValueError:
        return parse_date(value)
    else:
        return _now() + timedelta(seconds=value)


def serialize_date_delta(value):
    if isinstance(value, (float, int, long)):
        return str(int(value))
    else:
        return serialize_date(value)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
import shlex
from typing import Callable

from contrast_fireball import OtelAttributes, SpanType

from contrast.agent.policy.handlers import EventDict


def observe_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    cmd_param_location = event_dict.get("cmd", "")
    args_param_location = event_dict.get("args", "")
    shell_param_location = event_dict.get("shell", "")

    if not cmd_param_location and not args_param_location:
        raise ValueError(
            "Event must specify either 'cmd' or 'args' parameter location."
        )

    def observe_span_attrs(args: Mapping[str, object], result: object):
        shell = args.get(shell_param_location, False)
        command = args.get(cmd_param_location) or ""
        arguments = args.get(args_param_location)
        # The logic here comes from subprocess.Popen._execute_child, specifically
        # https://github.com/python/cpython/blob/c489934/Lib/subprocess.py#L1819-L1838
        if shell:
            if not command:
                command = "/bin/sh"
            command += " -c"
        if isinstance(arguments, str):
            arguments = shlex.quote(arguments)
        if isinstance(arguments, (list, tuple)):
            arguments = shlex.join(arguments)
        if command_string := f"{command} {arguments or ''}".strip():
            return {"cmd": command_string}
        return {}

    return SpanType.HostCommandExec, observe_span_attrs, None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.providers.hardcoded_value_rule import (
    HardcodedValueRule,
)


class HardcodedKey(HardcodedValueRule):
    # These are names, determined by the security team, that indicate a field
    # is likely to be a password or secret token of some sort.
    KEY_FIELD_NAMES = ["KEY", "AES", "DES", "IV", "SECRET"]

    # These are markers whose presence indicates that a field is more
    # likely to be a descriptor or requirement than an actual key.
    # We should ignore fields that contain them.
    NON_KEY_NAMES = ["CONTENT_CODES", "RESPONSE_CODES", "DIV", "SCHEMA"]

    @property
    def name(self):
        return "hardcoded-key"

    def is_name_valid(self, constant):
        constant_name_matches_denylisted_value = any(
            self.fuzzy_match(constant, self.KEY_FIELD_NAMES)
        )
        constant_name_matches_allowlisted_value = any(
            self.fuzzy_match(constant, self.NON_KEY_NAMES)
        )
        return (
            not constant_name_matches_denylisted_value
            or constant_name_matches_allowlisted_value
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Database adapter patch for psycopg2.
This module's cursor doesn't have an `executescript` method.
"""

import os
import sys
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)

PSYCOPG2 = "psycopg2"
VENDOR = "PostgreSQL"


class Psycopg2Patcher(dbapi2.Dbapi2Patcher):
    @fail_quietly("failed to get database inventory information")
    def extract_connection_attributes(self, connection, connect_args, connect_kwargs):
        """
        Record DB inventory for a Postgres connection.

        Here we make a good effort to find connection params. There are several ways
        that these can be set, in the following order of priority (using dbname as an
        example):
        - using the `connection_factory` kwarg
        - as a kwarg itself - `dbname` or the deprecated `database`
        - via the dbname parameter in the dsn string
        - with the PGDATABASE environment variable

        Newer versions of psycopg2 (v2.7, ~2017) support connection.get_dsn_parameters,
        which provides a dictionary of the parsed connection params - we're interested
        in `dbname`.

        For now, it's still possible for us to miss the dbname (i.e. an old version of
        psycopg2 using the dsn string only), but this is unlikely and it would only
        affect inventory.
        """
        dsn_params = getattr(connection, "get_dsn_parameters", dict)()
        dbname = (
            dsn_params.get("dbname")
            or connect_kwargs.get("dbname")
            or connect_kwargs.get("database")
            or os.environ.get("PGDATABASE", "unknown_database")
        )
        self.db_name = dbname
        super().extract_connection_attributes(connection, connect_args, connect_kwargs)


def instrument_psycopg2(psycopg2):
    dbapi2.instrument_adapter(
        psycopg2, VENDOR, Psycopg2Patcher, extra_cursors=[psycopg2.extensions.cursor]
    )


def register_patches():
    register_module_patcher(instrument_psycopg2, PSYCOPG2)


def reverse_patches():
    unregister_module_patcher(PSYCOPG2)
    if psycopg2 := sys.modules.get(PSYCOPG2):
        patch_manager.reverse_patches_by_owner(psycopg2)
        patch_manager.reverse_patches_by_owner(psycopg2.extensions.connection)
        patch_manager.reverse_patches_by_owner(psycopg2.extensions.cursor)
        patch_manager.reverse_patches_by_owner(psycopg2.extras.NamedTupleCursor)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Instrumentation for genshi globals function. Essentially what happens here is when
genshi calls the builtin eval function to eval the template code (and perform var lookup)
a custom mechanism is used in the genshi eval.py file. As a result we needed to patch
globals in order to inject our contrast__operator for modulo string operations.
"""

from contrast.utils.patch_utils import (
    build_and_apply_patch,
    wrap_and_watermark,
    register_module_patcher,
)
from contrast_rewriter import PropagationRewriter, add_dependency_to_globals


GENSHI_TEMPALTE_ENGINE_EVAL_MODULE = "genshi.template.eval"


def build_genshi_eval_globals_patch(orig_func, _, rule_applicator):
    del rule_applicator

    def genshi_eval_globals_patch(wrapped, instance, args, kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            for mod in PropagationRewriter.all_possible_injected_modules:
                add_dependency_to_globals(args[0], mod)

        return orig_func(args)

    return wrap_and_watermark(orig_func, genshi_eval_globals_patch)


def patch_genshi_globals(genshi_eval_module):
    build_and_apply_patch(
        genshi_eval_module.LookupBase,
        "globals",
        build_genshi_eval_globals_patch,
        (None,),
    )


def register_patches():
    register_module_patcher(patch_genshi_globals, GENSHI_TEMPALTE_ENGINE_EVAL_MODULE)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from dataclasses import dataclass
from socket import gethostname
import sys
import os
import resource
import subprocess

from threading import Thread

from contrast.reporting import Reporter
from contrast.utils.configuration_utils import get_platform
from contrast.reporting.teamserver_messages.server_inventory import ServerInventory
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


REPORT_RESOURCE_ID_THREAD_NAME = "ContrastReportServerInventory"


def report_server_runtime(
    reporting_client: Reporter, discover_cloud_resource_enabled: bool = True
) -> None:
    """
    Report runtime environment info to TS in a background thread.
    """
    logger.debug("Starting detect runtime environment background thread")
    Thread(
        name=REPORT_RESOURCE_ID_THREAD_NAME,
        daemon=True,
        target=_report_runtime_environment,
        args=(reporting_client, discover_cloud_resource_enabled),
    ).start()


@fail_loudly("Failed to report operating system runtime environment details.")
def _report_runtime_environment(
    reporting_client: Reporter, discover_cloud_resource_enabled
) -> None:
    provider = None
    resource_id = None

    if discover_cloud_resource_enabled:
        provider, resource_id = _detect_cloud_resource_id()

    memory_metrics = get_memory_metrics()

    reporting_client.add_message(
        ServerInventory(
            operating_system=get_platform(),
            runtime_path=sys.executable,
            runtime_version=sys.version.split(" ")[0],
            hostname=gethostname(),
            is_kubernetes=_is_kubernetes(),
            is_docker=_is_docker(),
            cloud_provider=provider,
            cloud_resource_id=resource_id,
            process_memory_limit_bytes=memory_metrics.process_memory_limit_bytes,
        )
    )


def _is_kubernetes():
    return "KUBERNETES_SERVICE_HOST" in os.environ


@fail_quietly()
def _is_docker():
    docker_env_file = "/.dockerenv"
    cgroup_file = "/proc/self/cgroup"

    if os.path.isfile(docker_env_file):
        return True

    try:
        with open(cgroup_file) as cgroup:
            return "/docker" in cgroup.read().strip()
    except FileNotFoundError:
        logger.debug("The file %s was not found", cgroup_file)

    return False


def _detect_cloud_resource_id():
    extract_from = {
        "aws": extract_aws_resource_id,
        "azure": extract_azure_resource_id,
        # "gcp": extract_gcp_resource_id, # TODO: PYT-3346
    }

    for provider, extract_id_func in extract_from.items():
        try:
            resource_id = extract_id_func()
        except ResourceIdError as e:
            logger.debug(
                "did not extract resource ID",
                provider=provider,
                exc_info=e,
            )
            continue

        logger.debug(
            "discovered resource ID",
            cloud_provider=provider,
            resource_id=resource_id,
        )
        return provider, resource_id
    return None, None


METADATA_ENDPOINT_ADDRESS = "169.254.169.254"
TOTAL_RETRIES = 5
RETRY_BACKOFF_FACTOR = 1  # wait 0s, 1s, 2s, 4s, 8s, ..., on request retry
NO_PROXIES = {"http": "", "https": ""}
TIMEOUT_SECONDS = 10


class ResourceIdError(Exception):
    pass


def extract_azure_resource_id() -> str:
    """
    Extract the resource ID from the Azure Instance Metadata Service.
    """
    import requests
    from requests.adapters import HTTPAdapter
    from urllib3.util import Retry
    from urllib3.exceptions import MaxRetryError

    # This endpoint comes from https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux#route-parameters
    RESOURCE_ID_ENDPOINT = f"http://{METADATA_ENDPOINT_ADDRESS}/metadata/instance/compute/resourceId?api-version=2023-07-01&format=text"

    # https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux#errors-and-debugging
    AZURE_RETRY_CODES = {410, 429, 500}

    with requests.Session() as session:
        retries = Retry(
            total=TOTAL_RETRIES,
            backoff_factor=RETRY_BACKOFF_FACTOR,
            status_forcelist=Retry.RETRY_AFTER_STATUS_CODES.union(AZURE_RETRY_CODES),
        )
        session.mount("http://", HTTPAdapter(max_retries=retries))

        try:
            response = session.get(
                RESOURCE_ID_ENDPOINT,
                proxies=NO_PROXIES,
                headers={"Metadata": "true"},
                timeout=TIMEOUT_SECONDS,
            )
            response.raise_for_status()
        except (MaxRetryError, requests.RequestException) as e:
            raise ResourceIdError("Failed to get Azure resource ID") from e
        return response.text


def extract_aws_resource_id() -> str:
    """
    Generate an AWS ARN for the EC2 instance using its Identity Document.

    See:
    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html
    """
    logger.debug("Attempting to extract AWS resource ID")

    import requests
    import requests.adapters
    import urllib3.util
    import urllib3.exceptions

    retry_config = urllib3.util.Retry(
        total=TOTAL_RETRIES,
        backoff_factor=RETRY_BACKOFF_FACTOR,
        # force a retry for status codes supporting Retry-After, even if the header is
        # missing
        status_forcelist=urllib3.util.Retry.RETRY_AFTER_STATUS_CODES,
    )

    with requests.Session() as session:
        session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=retry_config)
        )
        token = _get_aws_token(session)
        # a token is required for IMDSv2, but if we can't get one, still try IMDSv1 (no
        # token)
        headers = {"X-aws-ec2-metadata-token": token} if token else {}
        try:
            response = session.get(
                f"http://{METADATA_ENDPOINT_ADDRESS}/latest/dynamic/instance-identity/document",
                headers=headers,
                proxies=NO_PROXIES,
                timeout=TIMEOUT_SECONDS,
            )
            response.raise_for_status()
            identity_doc = response.json()
        except (
            urllib3.exceptions.MaxRetryError,
            requests.RequestException,
        ) as e:
            raise ResourceIdError("Failed to get AWS resource ID") from e

    logger.debug("Retrieved AWS identity document", aws_identity_doc=identity_doc)
    try:
        region = identity_doc["region"]
        account_id = identity_doc["accountId"]
        instance_id = identity_doc["instanceId"]
    except KeyError as e:
        raise ResourceIdError(
            "Failed to get AWS resource ID. "
            "Missing a required field in identity document."
        ) from e

    arn = ":".join(
        [
            "arn",
            "aws",
            "ec2",
            region,
            account_id,
            f"instance/{instance_id}",
        ]
    )
    return arn


def _get_aws_token(session) -> str | None:
    logger.debug("Retrieving AWS token")

    import requests
    import urllib3.exceptions

    try:
        token_response = session.put(
            f"http://{METADATA_ENDPOINT_ADDRESS}/latest/api/token",
            headers={
                "X-aws-ec2-metadata-token-ttl-seconds": "300",
            },
            proxies=NO_PROXIES,
            timeout=TIMEOUT_SECONDS,
        )
        token_response.raise_for_status()
        return token_response.text
    except (
        urllib3.exceptions.MaxRetryError,
        requests.RequestException,
    ) as e:
        logger.debug(
            "Unable to retrieve token for AWS IMDSv2 - proceeding without it",
            error=str(e),
        )
    return None


CGROUP_V1_PATH = "/sys/fs/cgroup/memory/memory.limit_in_bytes"
CGROUP_V2_PATH = "/sys/fs/cgroup/memory.max"


@dataclass
class MemoryMetrics:
    total_memory_limit_bytes: int | None
    """
    The effective total memory of the current machine, excluding swap memory. On bare
    metal, this is usually the amount of physical memory connected to the device (and
    visible to the OS). This can be further limited by cgroups (often in containers).
    """
    process_memory_limit_bytes: int | None
    """
    The maximum amount of memory useable by the current process. Must not exceed the
    total memory limit.
    """


def get_memory_metrics() -> MemoryMetrics:
    """
    Retrieve memory information for the current machine / process.
    """
    virtual_mem = _total_virtual_memory_bytes()
    cgroup_mem = _cgroup_memory_limit_bytes()
    _, process_hard_limit_mem = _process_rlimits_bytes()

    total_mem = virtual_mem
    if virtual_mem is not None and cgroup_mem is not None:
        total_mem = min(virtual_mem, cgroup_mem)

    process_mem = total_mem
    if total_mem is not None and process_hard_limit_mem is not None:
        process_mem = min(total_mem, process_hard_limit_mem)

    return MemoryMetrics(
        total_memory_limit_bytes=total_mem,
        process_memory_limit_bytes=process_mem,
    )


LINUX = sys.platform.startswith("linux")
MACOS = sys.platform.startswith("darwin")


@fail_quietly(return_value=None)
def _total_virtual_memory_bytes() -> int | None:
    """
    Get the total memory available on the machine in bytes. Based on the implementation
    of `virtual_memory` from https://github.com/giampaolo/psutil.
    """
    if LINUX:
        return _linux_virtual_memory()
    if MACOS:
        return _macos_virtual_memory()
    return None


def _linux_virtual_memory() -> int:
    """
    See https://github.com/giampaolo/psutil/blob/e7754af7400c0836a721befaba0f54c402c4ecb4/psutil/_pslinux.py#L351-L369
    """
    mems = {}
    # Buffering prevents several issues. See the larger comment in psutil.
    with open("/proc/meminfo", "rb", buffering=32 * 1024) as f:
        for line in f:
            fields = line.split()
            mems[fields[0]] = int(fields[1]) * 1024

    return mems[b"MemTotal:"]


def _macos_virtual_memory() -> int:
    """
    Get the total memory available on macOS in bytes.
    """
    output = subprocess.check_output(["sysctl", "-n", "hw.memsize"])
    return int(output.strip())


@fail_quietly(return_value=None)
def _cgroup_memory_limit_bytes() -> int | None:
    """
    Get the memory limit in bytes imposed by cgroups (used by Docker), or None if not
    found or unlimited.
    """
    for path in (CGROUP_V2_PATH, CGROUP_V1_PATH):
        if os.path.exists(path):
            with open(path) as f:
                val = f.read().strip()
            if not val.isdigit():
                return None
            limit = int(val)
            if limit < 0:
                return None
            return limit
    return None


@fail_quietly(return_value=(None, None))
def _process_rlimits_bytes() -> tuple[int | None, int | None]:
    """
    Get the current memory limits for this process in bytes, according to getrlimit.
    Values that cannot be found or are unlimited are reported as None.
    """
    soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_AS)
    return (
        soft_limit if soft_limit != resource.RLIM_INFINITY else None,
        hard_limit if hard_limit != resource.RLIM_INFINITY else None,
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.agent.assess.policy.patches import build_assess_method
from contrast.agent.policy import patch_manager, registry
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)


CODEC_MODULES_TO_PATCH = [
    "encodings.ascii",
    "encodings.latin_1",
    "encodings.raw_unicode_escape",
    "encodings.unicode_escape",
    "encodings.unicode_internal",
]


def build_codec_patch(module, method_name):
    module_name = module.__name__

    orig_method = getattr(module.Codec, method_name)
    patch_policy = registry.get_policy_by_name(f"{module_name}.Codec.{method_name}")

    return build_assess_method(orig_method, patch_policy)


def patch_codec_module(module):
    patch_manager.patch(module.Codec, "encode", build_codec_patch(module, "encode"))
    patch_manager.patch(module.Codec, "decode", build_codec_patch(module, "decode"))

    import encodings

    # Clear the encodings cache so that our patches are seen
    encodings._cache.clear()


def register_patches():
    for module in CODEC_MODULES_TO_PATCH:
        register_module_patcher(patch_codec_module, module)


def reverse_patches():
    for name in CODEC_MODULES_TO_PATCH:
        unregister_module_patcher(name)
        module = sys.modules.get(name)
        if module is not None:
            patch_manager.reverse_patches_by_owner(module.Codec)



"""
Expose zipp.Path as .zipfile.Path.

Includes everything else in ``zipfile`` to match future usage. Just
use:

>>> from zipp.compat.overlay import zipfile

in place of ``import zipfile``.

Relative imports are supported too.

>>> from zipp.compat.overlay.zipfile import ZipInfo

The ``zipfile`` object added to ``sys.modules`` needs to be
hashable (#126).

>>> _ = hash(sys.modules['zipp.compat.overlay.zipfile'])
"""

import importlib
import sys
import types

from contrast_vendor import zipp


class HashableNamespace(types.SimpleNamespace):
    def __hash__(self):
        return hash(tuple(vars(self)))


zipfile = HashableNamespace(**vars(importlib.import_module('zipfile')))
zipfile.Path = zipp.Path
zipfile._path = zipp

sys.modules[__name__ + '.zipfile'] = zipfile  # type: ignore[assignment]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import io
from operator import attrgetter

from aiohttp.web import BaseRequest
from contrast.agent.assess.policy.source_policy import (
    cs__apply_source,
    build_source_node,
)
from contrast.agent.request_context import RequestContext
from contrast.utils.decorators import fail_loudly


# the following sources are purposely omitted from source tracking:
# version
# method
# scheme
AIOHTTP_REQUEST_SOURCES = {
    "url._val": "URI",
    "rel_url._val": "URI",
    "host": "HEADER",
    "remote": "OTHER",
    "path_qs": "URI",
    "path": "URI",
    "raw_path": "URI",
    "query": "PARAMETER",
    "query_string": "QUERYSTRING",
    "headers": "HEADER",
    "raw_headers": "HEADER",
    "cookies": "COOKIE",
    "content_type": "HEADER",
    "charset": "HEADER",
    # body sources handled in policy
}


SOURCE_DICT = {
    "module": "aiohttp.web.BaseRequest",
    "instance_method": False,
    "target": "RETURN",
    "policy_patch": False,
}


@fail_loudly("Failed to convert aiohttp request to environ dict")
async def aiohttp_request_to_environ(request):
    """
    Converts an aiohttp.web.BaseRequest object to a valid WSGI environ dictionary
    """
    if ":" in request.host:
        host, port = request.host.split(":", 1)
    else:
        host = request.host
        port = "80"

    if request.content_type == "multipart/form-data":
        body = (
            b"Contrast placeholder - "
            b"we cannot safely capture aiohttp multipart form data"
        )
        real_body = False
    else:
        body = await request.read()
        real_body = True

    path = request.path
    query_string = request.query_string

    environ = {
        "REQUEST_METHOD": request.method,
        "SCRIPT_NAME": "",
        "SERVER_NAME": host,
        "SERVER_PORT": port,
        "PATH_INFO": path,
        "QUERY_STRING": query_string,
        "SERVER_PROTOCOL": f"HTTP/{request.version[0]}.{request.version[1]}",
        "REMOTE_ADDR": request.remote,
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": request.scheme,
        "wsgi.input": io.BytesIO(body),
        "wsgi.errors": io.BytesIO(),
        "wsgi.multithread": True,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
    }

    if request.content_length is not None:
        environ["CONTENT_LENGTH"] = (
            request.content_length if real_body else str(len(body))
        )
    if request.content_type is not None:
        environ["CONTENT_TYPE"] = request.content_type if real_body else "text/plain"

    for name, value in request.headers.items():
        if name.lower() in ("content-length", "content-type"):
            continue
        corrected_name = f"HTTP_{name.upper().replace('-', '_')}"
        if corrected_name in environ:
            value = environ[corrected_name] + "," + value
        environ[corrected_name] = value

    return environ


def track_aiohttp_request_sources(context: RequestContext, request: BaseRequest):
    """
    Explicitly tracks all relevant attributes of an aiohttp.web.BaseRequest object. This
    is similar to the environ/scope trackers.
    """
    if not context.assess_enabled:
        return

    for src_name, src_type in AIOHTTP_REQUEST_SOURCES.items():
        if attr := attrgetter(src_name)(request):
            node = build_source_node(SOURCE_DICT, src_name, src_type)
            cs__apply_source(
                context, node, attr, request, attr, (), {}, source_name=src_name
            )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from . import (
    cs_io,
    encodings_patch,
    re_patch,
    threading_patch,
    exec_and_eval,
    genshi_patch,
    lxml_patch,
    pathlib_patch,
    sys_patch,
    cgi_patch,
    concurrent_futures_thread_patch,
    str_new,
    operator,
    os_patch,
    jinja2_patch,
    # graphene_patch,  # TODO: PYT-3819 uncomment when ready
)

from .databases import (
    mysql_connector_patch,
    pymysql_patch,
    psycopg2_patch,
    sqlalchemy_patch,
    sqlite3_patch,
)

from .frameworks import (
    bottle_patches,
    django_patches,
    drf_patches,
    falcon_patches,
    pyramid_patch,
    starlette_patches,
    flask_and_quart_patches,
)

from .middleware import django, mod_wsgi, aiohttp
from .middleware.common import CommonMiddlewarePatch, AppInterfaceType

from . import chaining_patches

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

COMMON_PATCH_MODULES = (
    pathlib_patch,
    sqlalchemy_patch,
    sqlite3_patch,
    mysql_connector_patch,
    pymysql_patch,
    psycopg2_patch,
    concurrent_futures_thread_patch,
)


ASSESS_PATCH_MODULES = (
    operator,
    threading_patch,
    cs_io,
    encodings_patch,
    re_patch,
    exec_and_eval,
    genshi_patch,
    lxml_patch,
    pyramid_patch,
    django_patches,
    drf_patches,
    bottle_patches,
    flask_and_quart_patches,
    falcon_patches,
    sys_patch,
    cgi_patch,
    starlette_patches,
    str_new,
    os_patch,
    jinja2_patch,
    # graphene_patch,  # TODO: PYT-3819 uncomment when ready
)


MIDDLEWARE_PATCH_MODULES = (
    django,
    mod_wsgi,
    CommonMiddlewarePatch("flask"),
    CommonMiddlewarePatch("bottle"),
    CommonMiddlewarePatch(
        "pyramid.router",
        application_class_name="Router",
        framework_name="pyramid",
    ),
    CommonMiddlewarePatch(
        "fastapi",
        application_class_name="FastAPI",
        app_interface=AppInterfaceType.ASGI,
    ),
    CommonMiddlewarePatch(
        "starlette.applications",
        application_class_name="Starlette",
        framework_name="starlette",
        app_interface=AppInterfaceType.ASGI,
    ),
    aiohttp,
    CommonMiddlewarePatch(
        "falcon",
        application_class_name="App",
        framework_name="falcon",
        app_interface=AppInterfaceType.AUTO_DETECT,
    ),
    # NOTE: falcon.asgi.App is a subclass of falcon.App. Because of this, registering
    # this patch can cause multiple middleware initializations - for this case, it
    # doesn't cause any issues. We still need this patch, since the two classes do not
    # share any code for `__call__` (only for `__init__`)
    CommonMiddlewarePatch(
        "falcon.asgi",
        application_class_name="App",
        framework_name="falcon",
        app_interface=AppInterfaceType.ASGI,
    ),
    CommonMiddlewarePatch("quart", app_interface=AppInterfaceType.ASGI),
)


def _register_module_patches(module, patch_group):
    logger.debug("registering %s patches for %s", patch_group, module.__name__)

    try:
        module.register_patches()
    except Exception:
        logger.exception("failed to register patches for %s", module.__name__)


def register_common_monkeypatches():
    for module in COMMON_PATCH_MODULES:
        _register_module_patches(module, "common")


def register_assess_monkeypatches():
    for module in ASSESS_PATCH_MODULES:
        _register_module_patches(module, "assess")


def register_automatic_middleware_monkeypatches():
    for module in MIDDLEWARE_PATCH_MODULES:
        _register_module_patches(module, "middleware")


def register_chaining_monkeypatches():
    """
    Register patches to support chaining with other runners.

    Currently supports runners that use os.execl and PYTHONOPATH manipulation for
    sitecustomize loading.
    """
    chaining_patches.register_patches()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from typing import Any

import contrast_fireball
from contrast_vendor import webob

from contrast.api.user_input import InputType, DocumentType
from contrast.utils.string_utils import ensure_string, truncate
from contrast.utils.timer import now_ms
from contrast.agent.middlewares.route_coverage.common import (
    get_normalized_uri as strip_uri,
    get_url_parameters as find_parameters,
)

from contrast_vendor import structlog as logging
from contrast_vendor.webob.compat import cgi_FieldStorage
from contrast_vendor.webob.multidict import NestedMultiDict

logger = logging.getLogger("contrast")

Environ = dict[str, Any]


class Request(webob.BaseRequest):
    environ: Environ

    def __init__(self, environ: Environ):
        super().__init__(environ)

        self._document_type = None
        self._normalized_uri = None
        self._url_parameters = None
        self._multipart_headers = None

        # These fields are set by an ActivityMasker and will be used for reporting.
        self._masked = False
        self._masked_body: str | None = None
        self._masked_cookies = None
        self._masked_headers = None
        self._masked_params = None
        self._masked_query_string = None
        self._parsed_http = None

        self.timestamp_ms = now_ms()

    def to_fireball_assess_request(self) -> contrast_fireball.AssessRequest:
        return contrast_fireball.AssessRequest(
            body=truncate(self._reportable_body, length=4096),
            headers={
                k: ([v for v in (vs if isinstance(vs, list) else [vs])])
                for k, vs in self._reportable_headers.items()
            },
            method=ensure_string(self.method),
            parameters={
                k: ([v for v in (vs if isinstance(vs, list) else [vs])])
                for k, vs in self._reportable_params.items()
            },
            port=int(self.host_port),
            protocol=ensure_string(self.scheme),
            query_string=self._reportable_query_string,
            uri=ensure_string(self.path),
            standard_normalized_uri=ensure_string(self.get_normalized_uri()),
            version=self._get_http_version(),
        )

    @property
    def reportable_format(self):
        assert self._masked, "Request must be masked before reporting"

        r = self.to_fireball_assess_request()
        return {
            "body": r.body,
            # the WSGI environ supports only one value per request header. However
            # the server decides to handle multiple headers, we're guaranteed to
            # have only unique keys in request.request_headers (since we iterate
            # over webob's EnvironHeaders). Thus, each value list here is length-1.
            "headers": r.headers,
            "method": r.method,
            "parameters": r.parameters,
            "port": r.port,
            "protocol": r.protocol,
            "queryString": r.query_string,
            "uri": r.uri,
            "version": r.version,
        }

    @property
    def POST(self):
        """
        Return a MultiDict containing all the variables from a form
        request. Returns an empty dict-like object for non-form requests.

        Form requests are typically POST requests, however any other
        requests with an appropriate Content-Type are also supported.

        Swallows ValueErrors with "Invalid boundary in multipart form" message.
        """
        try:
            return super().POST
        except ValueError as e:
            if str(e).startswith("Invalid boundary in multipart form"):
                # This likely came form a malformed multipart/form-data request,
                # which is out of our control. Still log the error in case
                # there's another case we haven't considered.
                logger.debug("WARNING: failed to parse params", error=e)
                return {}
            raise

    @property
    def params(self):
        """
        Returns a NestedMultiDict of the query string and POST form parameters.

        Multipart form data is not included in this dict.
        """
        post_params = self.POST
        for key, values in post_params.items():
            if isinstance(values, cgi_FieldStorage):
                del post_params[key]
        return NestedMultiDict(self.GET, post_params)

    @property
    def _reportable_body(self):
        if self._masked_body is not None:
            return self._masked_body

        return ensure_string(self.body)

    @property
    def _reportable_cookies(self):
        if self._masked_cookies is not None:
            return self._masked_cookies

        return {ensure_string(k): ensure_string(v) for k, v in self.cookies.items()}

    @property
    def _reportable_headers(self):
        if self._masked_headers is not None:
            return self._masked_headers

        return {ensure_string(k): ensure_string(v) for k, v in self.headers.items()}

    @property
    def _reportable_params(self):
        if self._masked_params is not None:
            return self._masked_params

        return {ensure_string(k): ensure_string(v) for k, v in self.params.items()}

    @property
    def _reportable_query_string(self):
        if self._masked_query_string is not None:
            return self._masked_query_string

        return ensure_string(self.query_string)

    def get_multipart_headers(self):
        if self._multipart_headers is not None:
            return self._multipart_headers

        self._multipart_headers = {}
        for field_name, filename in self._get_file_info():
            self._multipart_headers[field_name] = filename
        return self._multipart_headers

    def get_normalized_uri(self) -> str:
        """
        A best-effort to remove client-specific information from the path.
        Example:
        /user/123456/page/12 -> /user/{n}/page/{n}
        """
        if self._normalized_uri is not None:
            return self._normalized_uri

        self._normalized_uri = strip_uri(self.path)
        return self._normalized_uri

    def get_url_parameters(self):
        """
        Returns the url parameters in a list.
        Example
        /user/123456/page/12 -> ["123456", "12"]
        """
        if self._url_parameters is not None:
            return self._url_parameters

        self._url_parameters = find_parameters(self.path)
        return self._url_parameters

    def get_body(self, as_text=False, errors="ignore"):
        """
        Get the raw request body in either bytes or as a decoded string.
        Note that we do not use webob's Request.text here, because we do not want this
        to fail in the event of a decoding error.

        :param as_text: Boolean indicating if we should attempt to return a decoded
            string
        :param errors: String indicating the unicode error handling strategy, passed to
            decode()
        :return: The request body as either bytes or a decoded string
        """
        if not as_text:
            return self.body

        return ensure_string(self.body, encoding=self.charset, errors=errors)

    def _get_http_version(self):
        """
        teamserver expects this field to be a string representing the HTTP version only.
        Using 'HTTP/1.1' is not acceptable and will cause vulnerabilities to be omitted
        from TS.
        """
        return self.http_version.split("/")[-1]

    def _get_document_type_from_header(self):
        """
        Returns the document type based on the content type header if present
        """
        content_type = self.content_type.lower()

        if not content_type:
            return None
        if "json" in content_type:
            return DocumentType.JSON
        if "xml" in content_type:
            return DocumentType.XML
        if "x-www-form-urlencoded" in content_type or "form-data" in content_type:
            return InputType.PARAMETER_VALUE

        return DocumentType.NORMAL

    def _get_document_type_from_body(self):
        str_body = self.get_body(as_text=True)

        if str_body.startswith("<?xml"):
            return DocumentType.XML
        if re.search(r"^\s*[{[]", str_body):
            return DocumentType.JSON

        return DocumentType.NORMAL

    def _get_document_type(self):
        if self._document_type is not None:
            return self._document_type

        self._document_type = self._get_document_type_from_header()
        if self._document_type is None:
            self._document_type = self._get_document_type_from_body()

        return self._document_type

    def _get_file_info(self):
        """
        Get the field names and filenames of uploaded files
        :return: list of tuples of (field_name, filename)
        """
        file_info = []
        for f in self.POST.values():
            if hasattr(f, "filename") and hasattr(f, "name"):
                file_info.append((f.name, f.filename))
                logger.debug("Found uploaded file: %s", f.filename)

        return file_info

    # from https://github.com/Contrast-Security-Inc/secobs-semantic-conventions/blob/main/docs/http/http-spans.md?plain=1#L121
    _OTEL_KNOWN_METHODS = {
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "DELETE",
        "CONNECT",
        "OPTIONS",
        "TRACE",
        "PATCH",
    }

    def get_otel_attributes(self) -> contrast_fireball.OtelAttributes:
        """
        Returns attributes following OpenTelemetry semantic conventions for HTTP spans.
        """
        attributes = {
            "url.path": self.path,
            "url.scheme": self.scheme,
        }

        # url.query is conditionally required if a query string is present
        if self.query_string:
            attributes["url.query"] = self._reportable_query_string

        # http.request.method_original is conditionally required if the method is not one of the known methods
        if self.method not in self._OTEL_KNOWN_METHODS:
            attributes["http.request.method"] = "_OTHER"
            attributes["http.request.method_original"] = self.method
        else:
            attributes["http.request.method"] = self.method

        return attributes
